import pandas as pd
from sklearn import metrics
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler
from sdmetrics.goal import Goal

from syntherela.metadata import drop_ids
from syntherela.metrics.base import DistanceBaseMetric, SingleTableMetric


class MaximumMeanDiscrepancy(DistanceBaseMetric, SingleTableMetric):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "MaximumMeanDiscrepancy"
        self.goal = Goal.MINIMIZE
        self.min_value = 0.0
        self.max_value = float("inf")

    @staticmethod
    def is_applicable(metadata):
        """
        Check if the table contains at least one column that is not an id.
        """
        for column_name in metadata["columns"].keys():
            if metadata["columns"][column_name]["sdtype"] == "numerical":
                return True
        return False

    @staticmethod
    def compute(original_table, sythetic_table, metadata, kernel="linear", **kwargs):
        """
        Code for MaximumMeanDiscrepancy metric modified from:
        Qian, Z., Cebere, B.-C., & van der Schaar, M. (2023).
        Synthcity: Facilitating innovative use cases of synthetic data in different data modalities.
        arXiv: https://arxiv.org/abs/2301.07573
        github: https://github.com/vanderschaarlab/synthcity/
        """
        orig = original_table.copy()
        synth = sythetic_table.copy()

        orig = drop_ids(orig, metadata)
        synth = drop_ids(synth, metadata)
        for col in orig.columns:
            if orig[col].dtype.name in ("object", "category"):
                orig.drop(col, axis=1, inplace=True)
                synth.drop(col, axis=1, inplace=True)
            elif "datetime" in str(orig[col].dtype):
                orig[col] = pd.to_numeric(orig[col])
                synth[col] = pd.to_numeric(synth[col])

        # standardize the values
        pipe = Pipeline(
            [
                ("imputer", SimpleImputer(strategy="mean")),
                ("scaler", StandardScaler()),
            ]
        )

        combined = pd.concat([orig, synth])
        pipe.fit(combined)
        orig = pipe.transform(orig)
        synth = pipe.transform(synth)

        if kernel == "linear":
            """
            MMD using linear kernel (i.e., k(x,y) = <x,y>)
            """
            delta = orig.mean(axis=0) - synth.mean(axis=0)
            # delta = delta_df.values

            score = delta.dot(delta.T)
        elif kernel == "rbf":
            """
            MMD using rbf (gaussian) kernel (i.e., k(x,y) = exp(-gamma * ||x-y||^2 / 2))
            """
            gamma = 1.0
            XX = metrics.pairwise.rbf_kernel(
                orig.reshape(len(original_table), -1),
                synth.reshape(len(original_table), -1),
                gamma,
            )
            YY = metrics.pairwise.rbf_kernel(
                synth.reshape(len(sythetic_table), -1),
                synth.reshape(len(sythetic_table), -1),
                gamma,
            )
            XY = metrics.pairwise.rbf_kernel(
                orig.reshape(len(original_table), -1),
                synth.reshape(len(sythetic_table), -1),
                gamma,
            )
            score = XX.mean() + YY.mean() - 2 * XY.mean()
        elif kernel == "polynomial":
            """
            MMD using polynomial kernel (i.e., k(x,y) = (gamma <X, Y> + coef0)^degree)
            """
            degree = 2
            gamma = 1
            coef0 = 0
            XX = metrics.pairwise.polynomial_kernel(
                orig.reshape(len(original_table), -1),
                orig.reshape(len(original_table), -1),
                degree,
                gamma,
                coef0,
            )
            YY = metrics.pairwise.polynomial_kernel(
                synth.numpy().reshape(len(sythetic_table), -1),
                synth.numpy().reshape(len(sythetic_table), -1),
                degree,
                gamma,
                coef0,
            )
            XY = metrics.pairwise.polynomial_kernel(
                orig.reshape(len(original_table), -1),
                synth.numpy().reshape(len(sythetic_table), -1),
                degree,
                gamma,
                coef0,
            )
            score = XX.mean() + YY.mean() - 2 * XY.mean()
        else:
            raise ValueError(f"Unsupported kernel {kernel}")

        return score.astype(float)



"""Metadata module."""

from rctgan.metadata import visualization
from rctgan.metadata.dataset import Metadata
from rctgan.metadata.errors import MetadataError, MetadataNotFittedError
from rctgan.metadata.table import Table

__all__ = (
    'Metadata',
    'MetadataError',
    'MetadataNotFittedError',
    'Table',
    'visualization'
)



# -*- coding: utf-8 -*-

"""Main rctgan module."""

import pickle
import warnings

from rctgan.errors import NotFittedError
from rctgan.relational.rctgan import RCTGAN
from rctgan.tabular.ctganpc import CTGAN, PC_CTGAN

class RCTGAN:
    """Automated generative modeling and sampling tool.

    Allows the users to generate synthetic data after creating generative models for their data.

    Args:
        model (type):
            Class of the model to use. Defaults to ``rctgan.relational.RCTGAN``.
        model_kwargs (dict):
            Keyword arguments to pass to the model. If no ``model`` is given,
            this defaults to using a ``RCTGAN`` with ``gaussian`` distribution
            and ``categorical_fuzzy`` categorical transformer.
    """

    _model_instance = None
    DEFAULT_MODEL = RCTGAN
    DEFAULT_MODEL_KWARGS = {
        'hyperparam': None,
        'metadata': None,
        'current_table': None
    }

    def __init__(self, model=None, model_kwargs=None):
        if model is None:
            model = model or self.DEFAULT_MODEL
            if model_kwargs is None:
                model_kwargs = self.DEFAULT_MODEL_KWARGS

        self._model = model
        self._model_kwargs = (model_kwargs or dict()).copy()

    def fit(self, metadata, tables=None, root_path=None):
        """Fit this sdvrctgan instance to the dataset data.

        Args:
            metadata (dict, str or Metadata):
                Metadata dict, path to the metadata JSON file or Metadata instance itself.
            tables (dict):
                Dictionary with the table names as key and ``pandas.DataFrame`` instances as
                values.  If ``None`` is given, the tables will be loaded from the paths
                indicated in ``metadata``. Defaults to ``None``.
            root_path (str or None):
                Path to the dataset directory. If ``None`` and metadata is
                a path, the metadata location is used. If ``None`` and
                metadata is a dict, the current working directory is used.
        """
        self._model_instance = self._model(metadata, root_path, **self._model_kwargs)
        self._model_instance.fit(tables)

    def sample(self, table_name=None, num_rows=None,
               sample_children=True, reset_primary_keys=False):
        """Generate synthetic data for one table or the entire dataset.

        If a ``table_name`` is given and ``sample_children`` is ``False``, a
        ``pandas.DataFrame`` with the values from the indicated table is returned.
        Otherwise, if ``sample_children`` is ``True``, a dictionary containing both
        the table and all its descendant tables is returned.

        If no ``table_name`` is given, the entire dataset is sampled and returned
        in a dictionary.

        If ``num_rows`` is given, the root tables of the dataset will contain the
        indicated number of rows. Otherwise, the number of rows will be the same
        as in the original dataset. Number of rows in the child tables cannot be
        controlled and always will depend on the values from the sampled parent
        tables.

        If ``reset_primary_keys`` is ``True``, the primary key generators will be
        reset.

        Args:
            table_name (str):
                Name of the table to sample from. If not passed, sample the entire
                dataset.
            num_rows (int):
                Amount of rows to sample. If ``None``, sample the same number of rows
                as there were in the original table.
            sample_children (bool):
                Whether or not sample child tables. Used only if ``table_name`` is
                given. Defaults to ``True``.
            reset_primary_keys (bool):
                Whether or not reset the primary keys generators. Defaults to ``False``.

        Returns:
            dict or pandas.DataFrame:
                - Returns a ``dict`` when ``sample_children`` is ``True`` with the sampled table
                  and child tables.
                - Returns a ``pandas.DataFrame`` when ``sample_children`` is ``False``.

        Raises:
            NotFittedError:
                A ``NotFittedError`` is raised when the ``sdvrctgan`` instance has not been fitted yet.
        """
        if self._model_instance is None:
            raise NotFittedError('sdvrctgan instance has not been fitted')

        return self._model_instance.sample(
            table_name,
            num_rows,
            sample_children=sample_children,
            reset_primary_keys=reset_primary_keys
        )

    def sample_all(self, num_rows=None, reset_primary_keys=False):
        """Sample the entire dataset.

        WARNING: This method is deprecated and will be removed in future relaeses. Please
        use the ``sample`` method instead.

        Args:
            num_rows (int):
                Number of rows to be sampled on the first parent tables. If ``None``,
                sample the same number of rows as in the original tables.
            reset_primary_keys (bool):
                Wheter or not reset the primary key generators. Defaults to ``False``.

        Returns:
            dict:
                Tables sampled.

        Raises:
            NotFittedError:
                A ``NotFittedError`` is raised when the ``sdvrctgan`` instance has not been fitted yet.
        """
        warnings.warn('`sample_all` is deprecated and will be removed soon. Please use `sample`',
                      DeprecationWarning)
        return self.sample(num_rows=num_rows, reset_primary_keys=reset_primary_keys)

    def save(self, path):
        """Save this sdvrctgan instance to the given path using pickle.

        Args:
            path (str):
                Path where the sdvrctgan instance will be serialized.
        """
        with open(path, 'wb') as output:
            pickle.dump(self, output)

    @classmethod
    def load(cls, path):
        """Load a sdvrctgan instance from a given path.

        Args:
            path (str):
                Path from which to load the sdvrctgan instance.
        """
        with open(path, 'rb') as f:
            return pickle.load(f)



import numpy as np
import pandas as pd
from sdmetrics.goal import Goal
from sdmetrics.utils import is_datetime

from syntherela.metrics.base import SingleColumnMetric, DistanceBaseMetric
from syntherela.metrics.single_column.distance.utils import get_histograms


_SQRT2 = np.sqrt(2)


class HellingerDistance(DistanceBaseMetric, SingleColumnMetric):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "HellingerDistance"
        self.goal = Goal.MINIMIZE
        self.min_value = 0.0
        self.max_value = 1

    @staticmethod
    def is_applicable(column_type):
        return column_type in ["categorical", "numerical", "datetime", "boolean"]

    @staticmethod
    def hellinger(p, q):
        return np.sqrt(np.sum((np.sqrt(p) - np.sqrt(q)) ** 2)) / _SQRT2

    @classmethod
    def compute(cls, orig_col, synth_col, bins, normalize_histograms=True, **kwargs):
        """Compute this metric.

        Args:
            real_data:
                The values from the real dataset.
            synthetic_data:
                The values from the synthetic dataset.
            bins:
                The bins to use for the histogram.

        Returns:
            Union[float, tuple[float]]:
                Metric output or outputs.
        """
        gt_freq, synth_freq = get_histograms(
            orig_col, synth_col, normalize=normalize_histograms, bins=bins
        )
        return cls.hellinger(gt_freq, synth_freq)

    def run(self, real_data, synthetic_data, **kwargs):
        if self.is_constant(real_data):
            return {
                "value": 0,
                "reference_ci": [0, 0],
                "bootstrap_mean": 0,
                "bootstrap_se": 0,
            }
        # check for datetime
        if is_datetime(real_data):
            real_data = pd.to_numeric(real_data, errors="coerce", downcast="integer")
            synthetic_data = pd.to_numeric(
                synthetic_data, errors="coerce", downcast="integer"
            )
        # compute bin values on the original data
        if real_data.dtype.name in ("object", "category", "bool"):
            bins = None
        else:
            real_data = real_data.dropna()
            synthetic_data = synthetic_data.dropna()
            bins = np.histogram_bin_edges(real_data)
        return super().run(real_data, synthetic_data, bins=bins, **kwargs)



import json

import numpy as np
import pandas as pd
from tqdm import tqdm
from scipy.stats import pearsonr


def read_utility_results(dataset, model="xgboost", run="1"):
    utility_model = []
    utility_feature = []
    utility_score = []

    dataset_name = dataset.split("-")[0].split("_")[0]
    with open(f"results/mle_{dataset_name}_{run}_0.json", "r") as f:
        results = json.load(f)[dataset_name]
    for method in [
        "SDV",
        "RCTGAN",
        "REALTABFORMER",
        "MOSTLYAI",
        "GRETEL_ACTGAN",
        "GRETEL_LSTM",
        "CLAVADDPM",
    ]:
        utility_model.append(results[method]["spearman_mean"])
        utility_feature.append(results[method]["feature_importance_spearman_mean"])
        utility_score.append(results[method][model]["synthetic_score"])
    utility_scores = np.array(utility_score)
    # Normalize the utility scores so they are comparable between datasets
    utility_scores = (utility_scores - np.min(utility_scores)) / (
        np.max(utility_scores) - np.min(utility_scores)
    )
    return utility_model, utility_feature, utility_scores.tolist()


def read_fidelity_results(
    dataset, model, metric, target_table=None, target_column=None, run="1"
):
    results = []
    for method in [
        "SDV",
        "RCTGAN",
        "REALTABFORMER",
        "MOSTLYAI",
        "GRETEL_ACTGAN",
        "GRETEL_LSTM",
        "CLAVADDPM",
    ]:
        with open(f"results/{run}/{dataset}_{method}_{run}_sample1.json", "r") as f:
            method_results = json.load(f)
        if "Aggregation" in metric:
            if dataset == "rossmann_subsampled":
                root_table = "store"
            elif dataset == "airbnb-simplified_subsampled":
                root_table = "users"
            elif dataset == "walmart_subsampled":
                root_table = "stores"
            results.append(
                method_results["multi_table_metrics"][f"{metric}-{model}"][root_table][
                    "accuracy"
                ]
            )
        elif "SingleColumn" in metric:
            results.append(
                method_results["single_column_metrics"][f"{metric}-{model}"][
                    target_table
                ][target_column]["accuracy"]
            )
        elif "SingleTable" in metric:
            results.append(
                method_results["single_table_metrics"][f"{metric}-{model}"][
                    target_table
                ]["accuracy"]
            )
    return results


fidelities_xgb = []
fidelities_lin = []
utilities = []
dataset_list = []
method_list = []
run_list = []
for dataset in [
    "rossmann_subsampled",
    "walmart_subsampled",
    "airbnb-simplified_subsampled",
]:
    if dataset == "rossmann_subsampled":
        target_table = "historical"
        target_column = "Customers"
    elif dataset == "airbnb-simplified_subsampled":
        target_table = "users"
        target_column = "country_destination"
    elif dataset == "walmart_subsampled":
        target_table = "depts"
        target_column = "Weekly_Sales"

    for run in range(3):
        run_id = str(run + 1)

        # Fidelity
        fidelity_xgb = read_fidelity_results(
            dataset, "XGBClassifier", "AggregationDetection", run=run_id
        )
        fidelity_lin = read_fidelity_results(
            dataset,
            "LogisticRegression",
            "AggregationDetection",
            target_table,
            run=run_id,
        )

        utility_scores = []
        for model in [
            "xgboost",
            "linear",
            "random_forest",
            "decision_tree",
            "knn",
            "svr",
            "mlp",
            "svc",
            "gaussian_nb",
        ]:
            if dataset == "airbnb-simplified_subsampled":
                if model == "svr":
                    continue
            else:
                if model == "svc" or model == "gaussian_nb":
                    continue

            # Utility
            try:
                utility_model, utility_feature, utility_score = read_utility_results(
                    dataset, model, run=run_id
                )
            except FileNotFoundError:
                print(f"Run {run_id} not found for {dataset}")
                utility_model, utility_feature, utility_score = read_utility_results(
                    dataset, model
                )

            utility_scores.append(utility_score)

        utility_score = np.array(utility_scores).mean(axis=0).tolist()
        fidelities_xgb.extend(fidelity_xgb)
        fidelities_lin.extend(fidelity_lin)
        utilities.extend(utility_score)
        dataset_list += [dataset] * len(fidelity_xgb)
        run_list += [run_id] * len(fidelity_xgb)

utilities = np.array(utilities)
fidelities_xgb = np.array(fidelities_xgb)
fidelities_lin = np.array(fidelities_lin)
dataset_list = np.array(dataset_list)
run_list = np.array(run_list)


def bootstrap_correlations(
    utilities, fidelities_xgb, fidelities_lin, dataset_list, datasets, m=10000
):
    xgb_corr = []
    lin_corr = []
    xgb_lin_diff = []
    for i in tqdm(range(m)):
        indices = np.where(np.isin(dataset_list, datasets))[0]
        np.random.seed(i)
        indices = np.random.choice(indices, len(indices), replace=True)
        utility_score = utilities[indices]
        fidelity_xgb = fidelities_xgb[indices]
        fidelity_lin = fidelities_lin[indices]

        corr_xgb = pearsonr(utility_score, fidelity_xgb)[0]
        corr_lin = pearsonr(utility_score, fidelity_lin)[0]

        xgb_corr.append(corr_xgb)
        lin_corr.append(corr_lin)
        xgb_lin_diff.append(corr_xgb - corr_lin)
    return xgb_corr, lin_corr, xgb_lin_diff


datasets = ["rossmann_subsampled", "walmart_subsampled", "airbnb-simplified_subsampled"]
results = {}
for dataset in datasets:
    xgb_corr, lin_corr, xgb_lin_diff = bootstrap_correlations(
        utilities, fidelities_xgb, fidelities_lin, dataset_list, [dataset]
    )
    results[dataset] = {
        "xgb_corr": xgb_corr,
        "lin_corr": lin_corr,
        "xgb_lin_diff": xgb_lin_diff,
    }
xgb_corr, lin_corr, xgb_lin_diff = bootstrap_correlations(
    utilities,
    fidelities_xgb,
    fidelities_lin,
    dataset_list,
    ["rossmann_subsampled", "walmart_subsampled", "airbnb-simplified_subsampled"],
)
results["all"] = {
    "xgb_corr": xgb_corr,
    "lin_corr": lin_corr,
    "xgb_lin_diff": xgb_lin_diff,
}


dataset_names = {
    "rossmann_subsampled": "Rossmann",
    "walmart_subsampled": "Walmart",
    "airbnb-simplified_subsampled": "Airbnb",
    "all": "Total",
}


def format_result(result, ci=False):
    import math

    se = np.std(result) / np.sqrt(len(result))
    mean = np.mean(result)
    se_digit = abs(int(math.log10(abs(se + 1e-8))))
    if ci:
        q = np.quantile(result, [0.05, 0.95])
        return (
            f"${mean.round(se_digit)} ({q[0].round(se_digit)}, {q[1].round(se_digit)})$"
        )

    return f"${mean.round(se_digit)}$".replace("-0.0 ", "0 ")


results_df = pd.DataFrame(
    columns=[
        "Dataset",
        "$\\rho(DDA_{XGB}, U)$",
        "$\\rho(LD, U)$",
        "$\\rho_{DDA} - \\rho_{LD}$",
    ]
)
for dataset, result in results.items():
    diff = np.array(result["xgb_corr"]) - np.array(result["lin_corr"])
    ddxgb = format_result(result["xgb_corr"])
    ld = format_result(result["lin_corr"])
    diff = format_result(diff, ci=True)
    results_df.loc[len(results_df)] = [dataset_names[dataset], ddxgb, ld, diff]

results_df.to_latex("results/tables/table6.tex", index=False, column_format="lccc")



import os
import argparse
from pathlib import Path

import relsyndgb
from synthcity.plugins import Plugins
from synthcity.benchmark import Benchmarks
from synthcity.plugins.core.constraints import Constraints
from synthcity.plugins.core.dataloader import GenericDataLoader
from relsyndgb.metadata import Metadata
from relsyndgb.data import load_tables, save_tables, remove_sdv_columns
import torch
from sklearn.impute import SimpleImputer
import pandas as pd
import numpy as np

Plugins(categories=["generic", "privacy"]).list()

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="airbnb-simplified_subsampled")
args.add_argument("--real_data_path", type=str, default="/Users/martinjurkovic/Documents/github_projects/relsyndgb/data/downloads")
args.add_argument("--synthetic_data_path", type=str, default="/Users/martinjurkovic/Documents/github_projects/relsyndgb/data/synthetic/")
args.add_argument("--full_sensitivity", type=bool, default=True)
args.add_argument("--retrain", type=bool, default=True)
args.add_argument("--run_id", type=str, default="1")
args = args.parse_args([])
dataset_name = args.dataset_name
full_sensitivity = args.full_sensitivity
retrain = args.retrain
real_data_path = args.real_data_path
synthetic_data_path = args.synthetic_data_path
run_id = args.run_id

MODEL_NAMES = ['bayesian_network', 'ctgan', 'ddpm', 'nflow', 'tvae']

metadata = Metadata().load_from_json(Path(real_data_path) / f'{dataset_name}/metadata.json')
real_data = load_tables(Path(real_data_path) / f'{dataset_name}', metadata)
real_data, metadata = remove_sdv_columns(real_data, metadata, validate=False)
metadata.validate_data(real_data)

for MODEL_NAME in MODEL_NAMES:
    synthetic_data = {}
    for table in real_data.keys():
        X = real_data[table]
        X_orig = X.copy()
        syn_model = Plugins().get(MODEL_NAME)
        syn_model.strict = False

        # check if any column is constant
        constant_columns = X.columns[X.nunique() == 1]
        if len(constant_columns) > 0:
            X = X.drop(columns=constant_columns)


        numeric_columns = X.select_dtypes(include='number').columns
        if len(numeric_columns) > 0:
            imp = SimpleImputer(strategy='mean')
            X[numeric_columns] = pd.DataFrame(imp.fit_transform(X[numeric_columns]), columns=numeric_columns)
        
        syn_model.fit(X)
        synthetic_data[table] = syn_model.generate(len(X))
        synthetic_data[table] = synthetic_data[table].dataframe()

        # append comstant columns
        for column in constant_columns:
            synthetic_data[table][column] = X_orig[column]

    save_data_path = Path(synthetic_data_path) / dataset_name / MODEL_NAME / run_id / 'sample1'
    save_tables(synthetic_data, save_data_path)


import os

import numpy as np
from matplotlib import rc
import matplotlib.pyplot as plt

from syntherela.visualisations.utils import get_x_tick_width_coef, get_dataset_info

rc("font", **{"family": "serif", "serif": ["Times"]})
rc("text", usetex=True)


def visualize_single_column_distance_metrics(
    granularity_level, metric_type, all_results, datasets, methods, **kwargs
):
    for dataset in datasets:
        base_metrics, base_metric_names, save_figs, save_figs_path, methods = (
            get_dataset_info(
                granularity_level, metric_type, all_results, dataset, methods, **kwargs
            )
        )

        for base_metric, base_metric_name in zip(base_metrics, base_metric_names):
            for table in all_results[dataset][list(all_results[dataset].keys())[0]][
                "single_column_metrics"
            ][base_metric].keys():
                try:
                    N = len(methods)  # number of methods
                    M = len(
                        all_results[dataset][list(all_results[dataset].keys())[0]][
                            "single_column_metrics"
                        ][base_metric][table].keys()
                    )  # number of columns

                    ind = np.arange(M)
                    width = 0.15

                    fig, ax = plt.subplots(figsize=(10, 7))
                    # set dpi
                    fig.dpi = 300

                    colors = plt.cm.viridis(
                        np.linspace(0.5, 1, N)
                    )  # create a color map

                    columns = all_results[dataset][
                        list(all_results[dataset].keys())[0]
                    ]["single_column_metrics"][base_metric][table].keys()
                    for j, method in enumerate(methods):
                        method_means = [
                            all_results[dataset][method]["single_column_metrics"][
                                base_metric
                            ][table][column]["value"]
                            for column in columns
                        ]
                        method_ses = [
                            all_results[dataset][method]["single_column_metrics"][
                                base_metric
                            ][table][column]["bootstrap_se"]
                            for column in columns
                        ]
                        ax.bar(
                            ind + width * j,
                            method_means,
                            width,
                            yerr=method_ses,
                            color=colors[j],
                        )

                    def get_x_tick_width_coef(N):
                        if N == 5:
                            return 2
                        elif N == 4:
                            return 1.5
                        elif N == 3:
                            return 1
                        elif N == 2:
                            return 0.5
                        else:
                            return 0

                    x_tick_width_coef = get_x_tick_width_coef(N)
                    ax.set_xticks(ind + x_tick_width_coef * width)
                    rotation = 20 if len(columns) > 6 else 0
                    ax.set_xticklabels(columns, fontsize=10, rotation=rotation)
                    # y_min = 0

                    # max_value = max([all_results[dataset][method]['single_column_metrics'][base_metric][table][column]["value"] for column in columns])
                    # y_max = max_value * 1.2
                    ax.set_ylim(0)
                    # ax.set_yticks(np.arange(y_min, 1.01, 0.1))
                    ax.set_ylabel("Metric Value")

                    # Create a legend
                    from matplotlib.lines import Line2D

                    custom_lines = [
                        Line2D([0], [0], color=colors[i], lw=4) for i in range(N)
                    ]
                    ax.legend(
                        custom_lines, methods, loc="upper center", ncol=N, fontsize=11
                    )

                    for j, column in enumerate(columns):
                        ci95 = all_results[dataset][method]["single_column_metrics"][
                            base_metric
                        ][table][column]["reference_ci"]
                        # ci95 = np.mean(ci95, axis=0)
                        ax.axhline(
                            y=ci95[1],
                            color="black",
                            linestyle="--",
                            linewidth=1,
                            xmin=j / len(columns),
                            xmax=(j + 1) / len(columns),
                        )
                        # ax.axhline(y=ci95[0], color='black', linestyle='--', linewidth=1,
                        #         xmin=j/len(columns),
                        #         xmax=(j+1)/len(columns))

                        if ci95[1] > ax.get_ylim()[1]:
                            y_max = ci95[1] * 1.1
                            ax.set_ylim(0, y_max)

                    # set title
                    plt.title(
                        f"{base_metric_name} for dataset {dataset}, table {table}"
                    )

                    if save_figs:
                        os.makedirs(save_figs_path, exist_ok=True)

                        plt.savefig(
                            f"{save_figs_path}/{dataset}_{table}_{base_metric}.png",
                            dpi=300,
                        )

                except Exception as e:
                    print(f"{base_metric_name} for dataset {dataset}, table {table}")
                    print(e)
                    pass


def visualize_single_column_detection_metrics(
    granularity_level, metric_type, all_results, datasets, methods, **kwargs
):
    for dataset in datasets:
        base_metrics, base_metric_names, save_figs, save_figs_path, methods = (
            get_dataset_info(
                granularity_level, metric_type, all_results, dataset, methods, **kwargs
            )
        )

        for base_metric, base_metric_name in zip(base_metrics, base_metric_names):
            for table in all_results[dataset][list(all_results[dataset].keys())[0]][
                "single_column_metrics"
            ][base_metric].keys():
                N = len(methods)  # number of methods
                M = len(
                    all_results[dataset][list(all_results[dataset].keys())[0]][
                        "single_column_metrics"
                    ][base_metric][table].keys()
                )  # number of columns

                ind = np.arange(M)
                width = 0.15

                fig, ax = plt.subplots(figsize=(10, 7))
                # set dpi
                fig.dpi = 300

                colors = plt.cm.viridis(np.linspace(0.5, 1, N))  # create a color map

                min_mean = 1

                for j, method in enumerate(methods):
                    if method not in all_results[dataset]:
                        continue
                    columns = all_results[dataset][method]["single_column_metrics"][
                        base_metric
                    ][table].keys()
                    method_means = [
                        all_results[dataset][method]["single_column_metrics"][
                            base_metric
                        ][table][column]["accuracy"]
                        for column in columns
                    ]
                    min_mean = min(min_mean, min(method_means))
                    method_ses = [
                        all_results[dataset][method]["single_column_metrics"][
                            base_metric
                        ][table][column]["SE"]
                        for column in columns
                    ]
                    baseline_means = np.array([0.5 for column in columns])
                    baseline_ses = np.array([0 for column in columns])
                    ax.bar(
                        ind + width * j,
                        method_means,
                        width,
                        yerr=method_ses,
                        color=colors[j],
                    )
                    # draw a line for baseline accuracy
                    ax.hlines(
                        baseline_means,
                        ind + width * j - width / 2,
                        ind + width * j + width / 2,
                        color="k",
                    )
                    ax.hlines(
                        baseline_means + 1.96 * baseline_ses,
                        ind + width * j - width / 2,
                        ind + width * j + width / 2,
                        color="k",
                        linestyle="--",
                    )
                    ax.hlines(
                        baseline_means - 1.96 * baseline_ses,
                        ind + width * j - width / 2,
                        ind + width * j + width / 2,
                        color="k",
                        linestyle="--",
                    )

                x_tick_width_coef = get_x_tick_width_coef(N)
                ax.set_xticks(ind + x_tick_width_coef * width)
                rotation = 20 if len(columns) > 6 else 0
                ax.set_xticklabels(columns, fontsize=10, rotation=rotation)
                y_min = 0.4 if min_mean > 0.4 else np.floor((min_mean - 0.1) * 10) / 10

                ax.set_ylim(y_min, 1.1)
                ax.set_ylabel("Metric Value")

                # Create a legend
                from matplotlib.lines import Line2D

                custom_lines = [
                    Line2D([0], [0], color=colors[i], lw=4) for i in range(N)
                ]
                ax.legend(
                    custom_lines, methods, loc="upper center", ncol=N, fontsize=11
                )

                ax.axhline(y=0.5, color="red", linestyle="--", linewidth=1)

                # set title
                plt.title(f"{base_metric_name} for dataset {dataset}, table {table}")

                if save_figs:
                    os.makedirs(save_figs_path, exist_ok=True)

                    plt.savefig(
                        f"{save_figs_path}/{dataset}_{table}_{base_metric}.png", dpi=300
                    )



import numpy as np
import pandas as pd
from scipy.stats import ks_2samp
from sdmetrics.utils import is_datetime
from sdmetrics.goal import Goal

from syntherela.metrics.base import SingleColumnMetric, StatisticalBaseMetric


class KolmogorovSmirnovTest(StatisticalBaseMetric, SingleColumnMetric):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "KolmogorovSmirnovTest"
        self.goal = Goal.MINIMIZE

    @staticmethod
    def is_applicable(column_type):
        return column_type == "numerical" or column_type == "datetime"

    def validate(self, column):
        column_dtype = column.dtypes
        if np.issubdtype(column_dtype, np.number) or np.issubdtype(
            column_dtype, np.datetime64
        ):
            return

        raise ValueError(
            f"{self.name} can only be applied to numerical columns, but column {column.name} is of type {column.dtype}"
        )

    @staticmethod
    def compute(real_data, synthetic_data):
        real_data = pd.Series(real_data).dropna()
        synthetic_data = pd.Series(synthetic_data).dropna()

        if is_datetime(real_data):
            real_data = pd.to_numeric(real_data)
            synthetic_data = pd.to_numeric(synthetic_data)

        statistic, p_val = ks_2samp(real_data, synthetic_data)

        return {"statistic": statistic, "p_val": p_val}



import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

from syntherela.visualisations.utils import get_bins

sns.set_theme()
# set colorblind color palette
sns.set_palette("colorblind")


def visualize_marginals(real_data, synthetic_data, metadata):
    for table in metadata.get_tables():
        table_meta = metadata.get_table_meta(table)
        num_non_id_columns = len(
            [
                column
                for column, column_info in table_meta["columns"].items()
                if column_info["sdtype"] != "id"
            ]
        )

        # round num_non_id_columns to the next multiple of 3
        if num_non_id_columns >= 3:
            num_non_id_columns = (
                num_non_id_columns + (3 - num_non_id_columns % 3)
                if num_non_id_columns % 3 != 0
                else num_non_id_columns
            )
        fig, axes = plt.subplots(
            max(num_non_id_columns // 3, 1),
            3,
            figsize=(15, 5 * (max(num_non_id_columns // 3, 1))),
        )
        i = 0
        for column, column_info in table_meta["columns"].items():
            if column_info["sdtype"] == "id":
                continue

            if num_non_id_columns <= 3:
                ax = axes[i]
            else:
                ax = axes[i // 3, i % 3]
            data = pd.DataFrame(
                pd.concat(
                    [real_data[table][column], synthetic_data[table][column]], axis=0
                )
            )
            data["Kind"] = ["Real"] * len(real_data[table]) + ["Synthetic"] * len(
                synthetic_data[table]
            )
            if (
                column_info["sdtype"] == "categorical"
                or column_info["sdtype"] == "boolean"
            ):
                data = data.astype("object")
                data.fillna("missing", inplace=True)
                sns.histplot(
                    data=data,
                    x=column,
                    hue="Kind",
                    multiple="dodge",
                    stat="density",
                    common_norm=False,
                    legend=True,
                    ax=ax,
                )
                # rotate x-axis labels
                if len(data[column].unique()) > 16:
                    ax.tick_params("x", labelrotation=90)
            elif (
                column_info["sdtype"] == "numerical"
                or column_info["sdtype"] == "datetime"
            ):
                data = data[data[column].notnull()]
                sns.kdeplot(
                    data=data,
                    x=column,
                    hue="Kind",
                    common_norm=False,
                    fill=False,
                    legend=True,
                    ax=ax,
                )
            ax.set_title(f"{table}.{column}")
            fig.tight_layout()
            fig.suptitle = f"{table}"
            i += 1
        if num_non_id_columns < 3:
            num_non_id_columns = 3
        for j in range(i, num_non_id_columns):
            if num_non_id_columns == 3:
                fig.delaxes(axes[j])
            else:
                fig.delaxes(axes[j // 3, j % 3])
    plt.show()


def visualize_bivariate_distributions(real_data, synthetic_data, metadata):
    for table in metadata.get_tables():
        table_meta = metadata.get_table_meta(table)
        non_id_columns = [
            column
            for column, column_info in table_meta["columns"].items()
            if column_info["sdtype"] != "id"
        ]
        pairs = [
            (non_id_columns[i], non_id_columns[j])
            for i in range(len(non_id_columns))
            for j in range(i + 1, len(non_id_columns))
        ]
        if len(pairs) == 0:
            continue

        fig, axes = plt.subplots(len(pairs), 2, figsize=(10, 5 * len(pairs)))
        for i, pair in enumerate(pairs):
            if len(pairs) == 1:
                ax1 = axes[0]
                ax2 = axes[1]
            else:
                ax1 = axes[i, 0]
                ax2 = axes[i, 1]
            data_real = pd.DataFrame(
                {pair[0]: real_data[table][pair[0]], pair[1]: real_data[table][pair[1]]}
            )
            data_synthetic = pd.DataFrame(
                {
                    pair[0]: synthetic_data[table][pair[0]],
                    pair[1]: synthetic_data[table][pair[1]],
                }
            )
            binsx = get_bins(data_real[pair[0]])
            binsy = get_bins(data_real[pair[1]])
            sns.histplot(
                data=data_real, x=pair[0], y=pair[1], ax=ax1, bins=(binsx, binsy)
            )
            ax1.set_title("Real")
            sns.histplot(
                data=data_synthetic, x=pair[0], y=pair[1], ax=ax2, bins=(binsx, binsy)
            )
            ax2.set_title("Synthetic")
            if (
                type(binsx) is int
                and binsx > 16
                or (type(binsx) is np.ndarray and len(binsx) > 16)
            ):
                ax1.tick_params("x", labelrotation=90)
                ax2.tick_params("x", labelrotation=90)
            xlim = ax1.get_xlim()
            ylim = ax1.get_ylim()
            ax2.set_xlim(xlim)
            ax2.set_ylim(ylim)
        fig.tight_layout()

    plt.show()


def visualize_parent_child_bivariates(real_data, synthetic_data, metadata):
    for table in metadata.get_tables():
        pairs = []
        table_meta = metadata.get_table_meta(table)
        non_id_columns = [
            column
            for column, column_info in table_meta["columns"].items()
            if column_info["sdtype"] != "id"
        ]
        for parent_table in metadata.get_parents(table):
            parent_table_meta = metadata.get_table_meta(parent_table)
            for column in parent_table_meta["columns"]:
                if parent_table_meta["columns"][column]["sdtype"] == "id":
                    continue
                for child_column in non_id_columns:
                    pairs.append(((parent_table, column), child_column))
        if len(pairs) == 0:
            continue

        fig, axes = plt.subplots(len(pairs), 2, figsize=(10, 5 * len(pairs)))
        for i, pair in enumerate(pairs):
            if len(pairs) == 1:
                ax1 = axes[0]
                ax2 = axes[1]
            else:
                ax1 = axes[i, 0]
                ax2 = axes[i, 1]
            parent_table, parent_column = pair[0]
            data_real = pd.DataFrame(
                {
                    pair[0]: real_data[parent_table][parent_column],
                    pair[1]: real_data[table][pair[1]],
                }
            )
            data_synthetic = pd.DataFrame(
                {
                    pair[0]: synthetic_data[parent_table][parent_column],
                    pair[1]: synthetic_data[table][pair[1]],
                }
            )
            binsx = get_bins(data_real[pair[0]])
            binsy = get_bins(data_real[pair[1]])
            sns.histplot(
                data=data_real, x=pair[0], y=pair[1], ax=ax1, bins=(binsx, binsy)
            )
            ax1.set_title("Real")
            ax1.set_xlabel(f"{parent_table}.{parent_column}")
            sns.histplot(
                data=data_synthetic, x=pair[0], y=pair[1], ax=ax2, bins=(binsx, binsy)
            )
            ax2.set_title("Synthetic")
            ax2.set_xlabel(f"{parent_table}.{parent_column}")
            if (
                type(binsx) is int
                and binsx > 16
                or (type(binsx) is np.ndarray and len(binsx) > 16)
            ):
                ax1.tick_params("x", labelrotation=90)
                ax2.tick_params("x", labelrotation=90)
            xlim = ax1.get_xlim()
            ylim = ax1.get_ylim()
            ax2.set_xlim(xlim)
            ax2.set_ylim(ylim)
        fig.tight_layout()



"""Personal Identifiable Information Anonymizer."""

import importlib
import inspect
import warnings
from copy import deepcopy

import faker
import numpy as np

from rctgan.rdt2.errors import Error
from rctgan.rdt2.transformers.base import BaseTransformer
from rctgan.rdt2.transformers.null import NullTransformer


class AnonymizedFaker(BaseTransformer):
    """Personal Identifiable Information Anonymizer using Faker.

    This transformer will drop a column and regenerate it with the previously specified
    ``Faker`` provider and ``function``. The transformer will also be able to handle nulls
    and regenerate null values if specified.

    Args:
        provider_name (str):
            The name of the provider in ``Faker``. If ``None`` the ``BaseProvider`` is used.
            Defaults to ``None``.
        function_name (str):
            The name of the function to use within the ``faker.provider``. Defaults to
            ``lexify``.
        function_kwargs (dict):
            Keyword args to pass into the ``function_name`` when being called.
        locales (list):
            List of localized providers to use instead of the global provider.
        missing_value_replacement (object or None):
            Indicate what to do with the null values. If an integer or float is given,
            replace them with the given value. If the strings ``'mean'`` or ``'mode'`` are
            given, replace them with the corresponding aggregation. If ``None`` is given,
            do not replace them. Defaults to ``None``.
        model_missing_values (bool):
            Whether to create a new column to indicate which values were null or not. The column
            will be created only if there are null values. If ``True``, create the new column if
            there are null values. If ``False``, do not create the new column even if there
            are null values. Defaults to ``False``.
    """

    DETERMINISTIC_TRANSFORM = False
    DETERMINISTIC_REVERSE = False
    INPUT_SDTYPE = 'pii'
    null_transformer = None

    @staticmethod
    def check_provider_function(provider_name, function_name):
        """Check that the provider and the function exist.

        Attempt to get the provider from ``faker.providers`` and then get the ``function``
        from the provider object. If one of them fails, it will raise an ``AttributeError``.

        Raises:
            ``AttributeError`` if the provider or the function is not found.
        """
        try:
            module = getattr(faker.providers, provider_name)
            if provider_name == 'BaseProvider':
                getattr(module, function_name)

            else:
                provider = getattr(module, 'Provider')
                getattr(provider, function_name)

        except AttributeError as exception:
            raise Error(
                f"The '{provider_name}' module does not contain a function named "
                f"'{function_name}'.\nRefer to the Faker docs to find the correct function: "
                'https://faker.readthedocs.io/en/master/providers.html'
            ) from exception

    def _check_locales(self):
        """Check if the locales exist for the provided provider."""
        locales = self.locales if isinstance(self.locales, list) else [self.locales]
        missed_locales = []
        for locale in locales:
            spec = importlib.util.find_spec(f'faker.providers.{self.provider_name}.{locale}')
            if spec is None:
                missed_locales.append(locale)

        if missed_locales:
            warnings.warn(
                f"Locales {missed_locales} do not support provider '{self.provider_name}' "
                f"and function '{self.function_name}'.\nIn place of these locales, 'en_US' will "
                'be used instead. Please refer to the localized provider docs for more '
                'information: https://faker.readthedocs.io/en/master/locales.html'
            )

    def __init__(self, provider_name=None, function_name=None, function_kwargs=None,
                 locales=None, missing_value_replacement=None, model_missing_values=False):
        self.data_length = None
        self.provider_name = provider_name if provider_name else 'BaseProvider'
        if self.provider_name != 'BaseProvider' and function_name is None:
            raise Error(
                'Please specify the function name to use from the '
                f"'{self.provider_name}' provider."
            )

        self.function_name = function_name if function_name else 'lexify'
        self.function_kwargs = deepcopy(function_kwargs) if function_kwargs else {}
        self.check_provider_function(self.provider_name, self.function_name)

        self.missing_value_replacement = missing_value_replacement
        self.model_missing_values = model_missing_values

        self.locales = locales
        self.faker = faker.Faker(locales)
        if self.locales:
            self._check_locales()

    def _function(self):
        """Return a callable ``faker`` function."""
        return getattr(self.faker, self.function_name)(**self.function_kwargs)

    def get_output_sdtypes(self):
        """Return the output sdtypes supported by the transformer.

        Returns:
            dict:
                Mapping from the transformed column names to supported sdtypes.
        """
        output_sdtypes = {}
        if self.null_transformer and self.null_transformer.models_missing_values():
            output_sdtypes['is_null'] = 'float'

        return self._add_prefix(output_sdtypes)

    def _fit(self, data):
        """Fit the transformer to the data.

        Args:
            data (pandas.Series):
                Data to fit to.
        """
        self.null_transformer = NullTransformer(
            self.missing_value_replacement,
            self.model_missing_values
        )
        self.null_transformer.fit(data)
        self.data_length = len(data)

    def _transform(self, data):
        """Drop the column and return ``null`` column if ``models_missing_values``.

        Args:
            data (pandas.Series):
                Data to transform.

        Returns:
            (numpy.ndarray or None):
                If ``self.model_missing_values`` is ``True`` then will return a ``numpy.ndarray``
                indicating which values should be ``nan``, else will return ``None``. In both
                scenarios the original column is being dropped.
        """
        if self.null_transformer and self.null_transformer.models_missing_values():
            return self.null_transformer.transform(data)[:, 1].astype(float)

        return None

    def _reverse_transform(self, data):
        """Generate new anonymized data using a ``faker.provider.function``.

        Args:
            data (pd.Series or numpy.ndarray):
                Data to transform.

        Returns:
            pandas.Series
        """
        reverse_transformed = np.array([
            self._function()
            for _ in range(self.data_length)
        ], dtype=object)

        if self.null_transformer.models_missing_values():
            reverse_transformed = np.column_stack((reverse_transformed, data))

        return self.null_transformer.reverse_transform(reverse_transformed)

    def __repr__(self):
        """Represent initialization of transformer as text.

        Returns:
            str:
                The name of the transformer followed by any non-default parameters.
        """
        class_name = self.__class__.__name__
        custom_args = []
        args = inspect.getfullargspec(self.__init__)
        keys = args.args[1:]
        defaults = dict(zip(keys, args.defaults))
        instanced = {key: getattr(self, key) for key in keys}

        defaults['function_name'] = None
        for arg, value in instanced.items():
            if value and defaults[arg] != value and value != 'BaseProvider':
                value = f"'{value}'" if isinstance(value, str) else value
                custom_args.append(f'{arg}={value}')

        args_string = ', '.join(custom_args)
        return f'{class_name}({args_string})'



import os
import warnings
from pathlib import Path
from datetime import datetime


from syntherela.report import Report
from syntherela.metadata import Metadata
from syntherela.data import load_tables, remove_sdv_columns
from syntherela.metrics.single_column.statistical import ChiSquareTest
from syntherela.metrics.single_table.distance import MaximumMeanDiscrepancy
from syntherela.visualisations.multi_table_visualisations import (
    visualize_multi_table,
)
from syntherela.visualisations.single_column_visualisations import (
    visualize_single_column_detection_metrics,
    visualize_single_column_distance_metrics,
)
from syntherela.visualisations.single_table_visualisations import (
    visualize_single_table_detection_metrics_per_table,
    visualize_single_table_distance_metrics,
)


class Benchmark:
    def __init__(
        self,
        real_data_dir,
        synthetic_data_dir,
        results_dir,
        benchmark_name,
        single_column_metrics=[
            ChiSquareTest(),
        ],
        single_table_metrics=[
            MaximumMeanDiscrepancy(),
        ],
        multi_table_metrics=[],
        methods=None,
        datasets=None,
        run_id=None,
        sample_id=None,
        validate_metadata=True,
    ):
        self.datasets = datasets
        self.run_id = str(run_id)
        self.sample_id = str(sample_id)
        self.validate_metadata = validate_metadata

        self.benchmark_name = (benchmark_name,)
        self.real_data_dir = Path(real_data_dir)
        self.synthetic_data_dir = Path(synthetic_data_dir)
        self.results_dir = Path(results_dir)

        if self.datasets is None:
            self.datasets = [
                d
                for d in os.listdir(self.synthetic_data_dir)
                if os.path.isdir(os.path.join(self.synthetic_data_dir, d))
            ]

        if methods is not None:
            # if self.methods is dict
            if isinstance(methods, dict):
                self.methods = methods
            if isinstance(methods, list):
                self.methods = {}
                for dataset_name in self.datasets:
                    self.methods[dataset_name] = methods
        else:
            self.methods = {}
            for dataset_name in self.datasets:
                if dataset_name not in self.methods:
                    self.methods[dataset_name] = [
                        d
                        for d in os.listdir(self.synthetic_data_dir / dataset_name)
                        if os.path.isdir(
                            os.path.join(self.synthetic_data_dir / dataset_name, d)
                        )
                    ]

        self.single_column_metrics = single_column_metrics
        self.single_table_metrics = single_table_metrics
        self.multi_table_metrics = multi_table_metrics
        self.benchmark_datetime = datetime.now()

        self.all_results = {}
        self.reports = {}

    def load_data(self, dataset_name, method_name):
        real_data_path = self.real_data_dir / dataset_name
        synthetic_data_path = self.synthetic_data_dir / dataset_name / method_name

        if self.run_id is not None:
            synthetic_data_path = synthetic_data_path / self.run_id
        if self.sample_id is not None:
            synthetic_data_path = synthetic_data_path / self.sample_id

        metadata = Metadata().load_from_json(real_data_path / "metadata.json")

        real_data = load_tables(real_data_path, metadata)
        synthetic_data = load_tables(synthetic_data_path, metadata)

        real_data, metadata = remove_sdv_columns(
            real_data, metadata, validate=self.validate_metadata
        )
        synthetic_data, metadata = remove_sdv_columns(
            synthetic_data,
            metadata,
            update_metadata=False,
            validate=self.validate_metadata,
        )

        return real_data, synthetic_data, metadata

    def run(self):
        for dataset_name in self.datasets:
            for method_name in self.methods[dataset_name]:
                try:
                    real_data, synthetic_data, metadata = self.load_data(
                        dataset_name, method_name
                    )

                    print(
                        f"Starting benchmark for {dataset_name}, method_name {method_name}"
                    )
                    report = Report(
                        real_data=real_data,
                        synthetic_data=synthetic_data,
                        metadata=metadata,
                        report_name=f"{self.benchmark_name}_{dataset_name}_{method_name}",
                        single_column_metrics=self.single_column_metrics,
                        single_table_metrics=self.single_table_metrics,
                        multi_table_metrics=self.multi_table_metrics,
                        validate_metadata=self.validate_metadata,
                    )

                    self.reports.setdefault(dataset_name, {})[method_name] = report

                    self.all_results.setdefault(dataset_name, {})[method_name] = (
                        report.generate()
                    )

                    file_name = self.build_file_name(dataset_name, method_name)

                    report.save_results(self.results_dir, file_name)
                except Exception as e:
                    print(
                        f"There was an error with dataset: {dataset_name}, method: {method_name}."
                    )
                    print(e)

    def read_results(self):
        for dataset_name in self.datasets:
            for method_name in self.methods[dataset_name]:
                file_name = self.build_file_name(dataset_name, method_name)
                real_data, synthetic_data, metadata = self.load_data(
                    dataset_name, method_name
                )
                temp_report = Report(
                    real_data,
                    synthetic_data,
                    metadata,
                    f"{dataset_name}_{method_name}",
                    validate_metadata=self.validate_metadata,
                ).load_from_json(self.results_dir / file_name)
                self.reports.setdefault(dataset_name, {})[method_name] = temp_report
                self.all_results.setdefault(dataset_name, {})[method_name] = (
                    temp_report.results
                )
        if not self.all_results:
            warnings.warn("No results found.")

    def build_file_name(self, dataset_name, method_name):
        file_name = f"{dataset_name}_{method_name}"
        if self.run_id is not None:
            file_name += f"_{self.run_id}"
            if self.sample_id is not None:
                file_name += f"_{self.sample_id}"
        file_name += ".json"
        return file_name

    def visualize_single_table_metrics(self, distance=True, detection=True, **kwargs):
        datasets = kwargs.pop("datasets", self.datasets)
        methods = kwargs.pop("methods", self.methods[datasets[0]])
        if distance:
            visualize_single_table_distance_metrics(
                granularity_level="single_table",
                metric_type="distance",
                all_results=self.all_results,
                datasets=datasets,
                methods=methods,
                **kwargs,
            )

        if detection:
            # visualize_single_table_detection_metrics_per_classifier(self.all_results, datasets, methods, **kwargs)
            visualize_single_table_detection_metrics_per_table(
                granularity_level="single_table",
                metric_type="detection",
                all_results=self.all_results,
                datasets=datasets,
                methods=methods,
                **kwargs,
            )

    def visualize_single_column_metrics(self, distance=True, detection=True, **kwargs):
        datasets = kwargs.get("datasets", self.datasets)
        methods = kwargs.get("methods", self.methods[datasets[0]])
        if distance:
            visualize_single_column_distance_metrics(
                granularity_level="single_column",
                metric_type="distance",
                all_results=self.all_results,
                datasets=datasets,
                methods=methods,
                **kwargs,
            )

        if detection:
            visualize_single_column_detection_metrics(
                granularity_level="single_column",
                metric_type="detection",
                all_results=self.all_results,
                datasets=datasets,
                methods=methods,
                **kwargs,
            )

    def visualize_multi_table_metrics(self, **kwargs):
        datasets = kwargs.get("datasets", self.datasets)
        methods = kwargs.get("methods", self.methods[datasets[0]])
        visualize_multi_table(self.all_results, datasets, methods, **kwargs)



"""Tools to evaluate the synthesized data."""

import pandas as pd
import sdmetrics

from rctgan.metadata.dataset import Metadata


def _validate_arguments(synthetic_data, real_data, metadata, root_path, table_name):
    """Validate arguments needed to compute descriptors values.

    If ``metadata`` is an instance of dict create the ``Metadata`` object.
    If ``metadata`` is ``None``, ``real_data`` has to be a ``pandas.DataFrane``.

    If ``real_data`` is ``None`` load all the tables and assert that ``synthetic_data`` is
    a ``dict``. Otherwise, ``real_data`` and ``synthetic_data`` must be of the same type.

    If ``synthetic_data`` is not a ``dict``, create a dictionary using the ``table_name``.

    Assert that ``synthetic_data`` and ``real_data`` must have the same tables.

    Args:
        synthetic_data (dict or pandas.DataFrame):
            Synthesized data.
        real_data (dict, pandas.DataFrame or None):
            Real data.
        metadata (str, dict, Metadata or None):
            Metadata instance or details needed to build it.
        root_path (str):
            Path to the metadata file.
        table_name (str):
            Table name used to prepare the metadata object, real_data and synthetic_data dict.

    Returns:
        tuple (dict, dict, Metadata):
            Processed tables and Metadata oject.
    """
    if isinstance(metadata, dict):
        metadata = Metadata(metadata, root_path)
    elif metadata is None:
        if not isinstance(real_data, pd.DataFrame):
            raise TypeError('If metadata is None, `real_data` has to be a DataFrame')

        metadata = Metadata()
        metadata.add_table(table_name, data=real_data)

    if real_data is None:
        real_data = metadata.load_tables()
        if not isinstance(synthetic_data, dict):
            raise TypeError('If `real_data` is `None`, `synthetic_data` must be a dict')

    elif not isinstance(synthetic_data, type(real_data)):
        raise TypeError('`real_data` and `synthetic_data` must be of the same type')

    if not isinstance(synthetic_data, dict):
        synthetic_data = {table_name: synthetic_data}

    if not isinstance(real_data, dict):
        real_data = {table_name: real_data}

    if not set(real_data.keys()) == set(synthetic_data.keys()):
        raise ValueError('real_data and synthetic dataset must have the same tables')

    if len(real_data.keys()) < len(metadata.get_tables()):
        meta_dict = {
            table: metadata.get_table_meta(table)
            for table in real_data.keys()
        }
        metadata = Metadata({'tables': meta_dict})

    return synthetic_data, real_data, metadata.to_dict()


def _select_metrics(synthetic_data, metrics):
    if isinstance(synthetic_data, dict):
        modality = 'multi-table'
        metric_classes = sdmetrics.multi_table.MultiTableMetric.get_subclasses()
    else:
        modality = 'single-table'
        metric_classes = sdmetrics.single_table.SingleTableMetric.get_subclasses()

    if metrics is None:
        return metric_classes, modality

    final_metrics = {}
    for metric in metrics:
        if isinstance(metric, str):
            try:
                final_metrics[metric] = metric_classes[metric]
            except KeyError:
                raise ValueError(f'Unknown {modality} metric: {metric}')

    return final_metrics, modality


def evaluate(synthetic_data, real_data=None, metadata=None, root_path=None,
             table_name=None, metrics=None, aggregate=True):
    """Apply multiple metrics at once.

    Args:
        synthetic_data (dict[str, pandas.DataFrame] or pandas.DataFrame):
            Map of names and tables of synthesized data. When evaluating a single table,
            a single ``pandas.DataFrame`` can be passed alone.
        real_data (dict[str, pandas.DataFrame] or pandas.DataFrame):
            Map of names and tables of real data. When evaluating a single table,
            a single ``pandas.DataFrame`` can be passed alone.
        metadata (str, dict, Metadata or None):
            Metadata instance or details needed to build it.
        root_path (str):
            Relative path to find the metadata.json file when needed.
        metrics (list[str]):
            List of metric names to apply.
        table_name (str):
            Table name to be evaluated, only used when ``synthetic_data`` is a
            ``pandas.DataFrame`` and ``real_data`` is ``None``.
        aggregate (bool):
            If ``get_report`` is ``False``, whether to compute the mean of all the normalized
            scores to return a single float value or return a ``dict`` containing the score
            that each metric obtained. Defaults to ``True``.

    Return:
        float or sdmetrics.MetricsReport
    """
    metrics, modality = _select_metrics(synthetic_data, metrics)

    synthetic_data, real_data, metadata = _validate_arguments(
        synthetic_data, real_data, metadata, root_path, table_name)

    if modality == 'single-table':
        table = list(metadata['tables'].keys())[0]
        metadata = metadata['tables'][table]
        real_data = real_data[table]
        synthetic_data = synthetic_data[table]

    scores = sdmetrics.compute_metrics(metrics, real_data, synthetic_data, metadata=metadata)

    if aggregate:
        return scores.normalized_score.mean()

    return scores



import os
import json
from pathlib import Path
from datetime import datetime

from tqdm import tqdm

from syntherela.utils import NpEncoder
from syntherela.metrics.single_column.statistical import ChiSquareTest
from syntherela.metrics.single_table.distance import MaximumMeanDiscrepancy
from syntherela.visualisations.distribution_visualisations import (
    visualize_bivariate_distributions,
    visualize_marginals,
    visualize_parent_child_bivariates,
)


class Report:
    def __init__(
        self,
        real_data,
        synthetic_data,
        metadata,
        report_name,
        single_column_metrics=[
            ChiSquareTest(),
        ],
        single_table_metrics=[
            MaximumMeanDiscrepancy(),
        ],
        multi_table_metrics=[],
        validate_metadata=True,
    ):
        if validate_metadata:
            metadata.validate_data(real_data)
            metadata.validate_data(synthetic_data)
        self.real_data = real_data
        self.synthetic_data = synthetic_data

        # reorder synthetic data columns to match real data
        for table in metadata.get_tables():
            self.synthetic_data[table] = self.synthetic_data[table][
                self.real_data[table].columns
            ]
            assert (
                self.real_data[table].columns == self.synthetic_data[table].columns
            ).all(), (
                f"Columns in real and synthetic data do not match for table {table}"
            )
        self.metadata = metadata
        self.report_name = report_name
        self.single_column_metrics = single_column_metrics
        self.single_table_metrics = single_table_metrics
        self.multi_table_metrics = multi_table_metrics
        self.report_datetime = datetime.now()
        self.results = {
            "single_column_metrics": {},
            "single_table_metrics": {},
            "multi_table_metrics": {},
        }

    def generate(self):
        """
        Generate the report.
        """
        column_count = sum(
            [len(self.real_data[table].columns) for table in self.metadata.get_tables()]
        )
        table_count = len(self.metadata.get_tables())

        # single_column_metrics
        if len(self.single_column_metrics) == 0:
            print("No single column metrics to run. Skipping.")
        else:
            with tqdm(
                total=len(self.single_column_metrics) * column_count,
                desc="Running Single Column Metrics",
            ) as pbar:
                for table in self.metadata.get_tables():
                    for metric in self.single_column_metrics:
                        for column, column_info in self.metadata.tables[
                            table
                        ].columns.items():
                            if not metric.is_applicable(column_info["sdtype"]):
                                pbar.update(1)
                                continue
                            try:
                                self.results["single_column_metrics"].setdefault(
                                    metric.name, {}
                                ).setdefault(table, {})[column] = metric.run(
                                    self.real_data[table][column],
                                    self.synthetic_data[table][column],
                                    metadata=self.metadata.to_dict()["tables"][table][
                                        "columns"
                                    ][column],
                                )
                            except Exception as e:
                                print(
                                    f"There was a problem with metric {metric.name}, table {table}, column {column}"
                                )
                                print(e)
                            pbar.update(1)

        # single_table_metrics
        if len(self.single_table_metrics) == 0:
            print("No single table metrics to run. Skipping.")
        else:
            with tqdm(
                total=len(self.single_table_metrics) * table_count,
                desc="Running Single Table Metrics",
            ) as pbar:
                for table in self.metadata.get_tables():
                    for metric in self.single_table_metrics:
                        if not metric.is_applicable(
                            self.metadata.to_dict()["tables"][table]
                        ):
                            pbar.update(1)
                            continue
                        try:
                            self.results["single_table_metrics"].setdefault(
                                metric.name, {}
                            )[table] = metric.run(
                                self.real_data[table],
                                self.synthetic_data[table],
                                metadata=self.metadata.to_dict()["tables"][table],
                            )
                        except Exception as e:
                            print(
                                f"There was a problem with metric {metric.name}, table {table}"
                            )
                            print(e)
                        pbar.update(1)

        # multi_table_metrics
        if len(self.multi_table_metrics) == 0:
            print("No multi table metrics to run. Skipping.")
        else:
            for metric in tqdm(
                self.multi_table_metrics, desc="Running Multi Table Metrics"
            ):
                try:
                    self.results["multi_table_metrics"][metric.name] = metric.run(
                        self.real_data,
                        self.synthetic_data,
                        metadata=self.metadata,
                    )
                except Exception as e:
                    print(f"There was a problem with metric {metric.name}")
                    print(e)

        self.report_datetime = datetime.now()

        return self.results

    def load_from_json(self, path):
        """
        Read the results from a file.
        """
        path = Path(path)
        with open(path, "r") as f:
            self.results = json.load(f)
        return self

    def print_results(self):
        """
        Print the results.
        """
        print(json.dumps(self.results, sort_keys=True, indent=4, cls=NpEncoder))

    def save_results(self, path, filename=None):
        """
        Save the results to a file.
        """
        path = Path(path)

        if filename is None:
            filename = (
                f"{self.report_name}_{self.report_datetime.strftime('%Y_%m_%d')}.json"
            )

        path = path / filename

        os.makedirs(os.path.dirname(path), exist_ok=True)

        with open(path, "w") as f:
            json.dump(self.results, f, sort_keys=True, indent=4, cls=NpEncoder)

    def visualize_distributions(
        self, marginals=True, bivariate=True, parent_child_bivariate=True
    ):
        """
        Visualize the distributions.
        """
        if marginals:
            visualize_marginals(self.real_data, self.synthetic_data, self.metadata)
        if bivariate:
            visualize_bivariate_distributions(
                self.real_data, self.synthetic_data, self.metadata
            )
        if parent_child_bivariate:
            visualize_parent_child_bivariates(
                self.real_data, self.synthetic_data, self.metadata
            )

    def get_metric_instance(self, metric_name):
        """
        Get the metric instance.
        """
        for metric in (
            self.single_column_metrics
            + self.single_table_metrics
            + self.multi_table_metrics
        ):
            if metric.name == metric_name:
                return metric
        raise ValueError(
            f'Metric with name "{metric_name}" not found in the report. Available metrics: {[metric.name for metric in self.single_column_metrics + self.single_table_metrics + self.multi_table_metrics]}'
        )



"""Metrics to evaluate quality of Synthetic Data.

This subpackage exists only to enable importing sdmetrics as part of sdv.
"""

from rctgan.metrics import relational, tabular, timeseries

__all__ = [
    'relational',
    'tabular',
    'timeseries',
]



"""Functions to evaluate and test the performance of the rdt2 Transformers."""

from rctgan.rdt2.performance import profiling
from rctgan.rdt2.performance.performance import evaluate_transformer_performance

__all__ = [
    'evaluate_transformer_performance',
    'profiling',
]



from syntherela.metadata import Metadata
from syntherela.metrics.multi_table.detection.denormalized_detection import (
    DenormalizedDetection,
)


class ParentChildDetection(DenormalizedDetection):
    @staticmethod
    def is_applicable(metadata: Metadata, table1: str, table2: str):
        """
        Check if the table contains at least one column that is not an id.
        And if the table has a relationship with another table.
        """
        nonid1 = False
        table_metadata = metadata.tables[table1].to_dict()
        for column_name in table_metadata["columns"].keys():
            if table_metadata["columns"][column_name]["sdtype"] != "id":
                nonid1 = True
                break
        nonid2 = False
        table_metadata = metadata.tables[table2].to_dict()
        for column_name in table_metadata["columns"].keys():
            if table_metadata["columns"][column_name]["sdtype"] != "id":
                nonid2 = True
                break
        return nonid1 and nonid2

    def prepare_data(
        self,
        real_data,
        synthetic_data,
        metadata,
        parent_table,
        child_table,
        pair_metadata,
    ):
        real_data_pair = {
            parent_table: real_data[parent_table],
            child_table: real_data[child_table],
        }
        synthetic_data_pair = {
            parent_table: synthetic_data[parent_table],
            child_table: synthetic_data[child_table],
        }
        return super().prepare_data(real_data_pair, synthetic_data_pair, pair_metadata)

    def run(self, real_data: dict, synthetic_data: dict, metadata: Metadata, **kwargs):
        results = {}
        for relationship in metadata.relationships:
            child_table = relationship["child_table_name"]
            child_fk = relationship["child_foreign_key"]
            parent_table = relationship["parent_table_name"]
            if not self.is_applicable(metadata, parent_table, child_table):
                continue
            pair_meta = metadata.to_dict()
            for table in metadata.get_tables():
                if table != parent_table and table != child_table:
                    pair_meta["tables"].pop(table)
            pair_meta["relationships"] = [relationship]
            pair_metadata = Metadata.load_from_dict(pair_meta)
            results[f"{parent_table}_{child_table}_{child_fk}"] = super().run(
                real_data=real_data,
                synthetic_data=synthetic_data,
                metadata=metadata,
                parent_table=parent_table,
                child_table=child_table,
                pair_metadata=pair_metadata,
            )
        return results



"""Table constraints.

This module contains constraints that are evaluated within a single table,
and which can affect one or more columns at a time, as well as one or more
rows.

Currently implemented constraints are:

    * CustomConstraint: Simple constraint to be set up by passing the python
      functions that will be used for transformation, reverse transformation
      and validation.
    * UniqueCombinations: Ensure that the combinations of values
      across several columns are the same after sampling.
    * GreaterThan: Ensure that the value in one column is always greater than
      the value in another column.
    * Positive: Ensure that the values in given columns are always positive.
    * Negative: Ensure that the values in given columns are always negative.
    * ColumnFormula: Compute the value of a column based on applying a formula
      on the other columns of the table.
    * Between: Ensure that the value in one column is always between the values
      of two other columns/scalars.
    * OneHotEncoding: Ensure the rows of the specified columns are one hot encoded.
"""

import operator
import uuid

import numpy as np
import pandas as pd

from rctgan.constraints.base import Constraint, import_object
from rctgan.constraints.utils import is_datetime_type


class CustomConstraint(Constraint):
    """Custom Constraint Class.

    This class simply takes the ``transform``, ``reverse_transform``
    and ``is_valid`` methods as optional arguments, so users can
    pass custom functions for each one of them.

    Args:
        transform (callable):
            Function to replace the ``transform`` method.
        reverse_transform (callable):
            Function to replace the ``reverse_transform`` method.
        is_valid (callable):
            Function to replace the ``is_valid`` method.
    """

    def _run(self, function, table_data, reverse=False):
        table_data = table_data.copy()
        if self._columns:
            if reverse:
                columns = reversed(self._columns)
            else:
                columns = self._columns

            for column in columns:
                try:
                    table_data = function(table_data, column)
                except TypeError:
                    table_data[column] = function(table_data[column])

        else:
            table_data = function(table_data)

        return table_data

    def _run_transform(self, table_data):
        return self._run(self._transform, table_data)

    def _run_reverse_transform(self, table_data):
        return self._run(self._reverse_transform, table_data, reverse=True)

    def _run_is_valid(self, table_data):
        if self._columns:
            try:
                valid = [self._is_valid(table_data, column) for column in self._columns]
            except TypeError:
                valid = [self._is_valid(table_data[column]) for column in self._columns]

            return np.logical_and.reduce(valid)

        return self._is_valid(table_data)

    def __init__(self, columns=None, transform=None, reverse_transform=None, is_valid=None):
        if isinstance(columns, str):
            self._columns = [columns]
        else:
            self._columns = columns

        self.fit_columns_model = False
        if transform is not None:
            self._transform = import_object(transform)
            self.transform = self._run_transform

        if reverse_transform is not None:
            self._reverse_transform = import_object(reverse_transform)
            self.reverse_transform = self._run_reverse_transform

        if is_valid is not None:
            self._is_valid = import_object(is_valid)
            self.is_valid = self._run_is_valid


class UniqueCombinations(Constraint):
    """Ensure that the combinations across multiple colums stay unique.

    One simple example of this constraint can be found in a table that
    contains the columns `country` and `city`, where each country can
    have multiple cities and the same city name can even be found in
    multiple countries, but some combinations of country/city would
    produce invalid results.

    This constraint would ensure that the combinations of country/city
    found in the sampled data always stay within the combinations previously
    seen during training.

    Args:
        columns (list[str]):
            Names of the columns that need to produce unique combinations. Must
            contain at least two columns.
        handling_strategy (str):
            How this Constraint should be handled, which can be ``transform``,
            ``reject_sampling`` or ``all``. Defaults to ``transform``.
        fit_columns_model (bool):
            If False, reject sampling will be used to handle conditional sampling.
            Otherwise, a model will be trained and used to sample other columns
            based on the conditioned column. Defaults to False.
    """

    _separator = None
    _joint_column = None
    _combinations_to_uuids = None
    _uuids_to_combinations = None

    def __init__(self, columns, handling_strategy='transform', fit_columns_model=False):
        if len(columns) < 2:
            raise ValueError('UniqueCombinations requires at least two constraint columns.')

        self._columns = columns
        self.constraint_columns = tuple(columns)
        self.rebuild_columns = tuple(columns)
        super().__init__(handling_strategy=handling_strategy,
                         fit_columns_model=fit_columns_model)

    def _fit(self, table_data):
        """Fit this Constraint to the data.

        The fit process consists on:

            - Finding a separator that works for the
              current data by iteratively adding `#` to it.
            - Generating the joint column name by concatenating
              the names of ``self._columns`` with the separator.
            - Generating a mapping of the unique combinations
              to a unique identifier.

        Args:
            table_data (pandas.DataFrame):
                Table data.
        """
        self._separator = '#'
        while self._separator.join(self._columns) in table_data:
            self._separator += '#'

        self._joint_column = self._separator.join(self._columns)
        self._combinations = table_data[self._columns].drop_duplicates().copy()
        self._combinations_to_uuids = {}
        self._uuids_to_combinations = {}
        for combination in self._combinations.itertuples(index=False, name=None):
            uuid_str = str(uuid.uuid4())
            self._combinations_to_uuids[combination] = uuid_str
            self._uuids_to_combinations[uuid_str] = combination

    def is_valid(self, table_data):
        """Say whether the column values are within the original combinations.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.Series:
                Whether each row is valid.
        """
        merged = table_data.merge(
            self._combinations,
            how='left',
            on=self._columns,
            indicator=self._joint_column
        )
        return merged[self._joint_column] == 'both'

    def _transform(self, table_data):
        """Transform the table data.

        The transformation consist on removing all the ``self._columns`` from
        the dataframe, and replacing them with a unique identifier that maps to
        that unique combination of column values under the previously computed
        combined column name.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Transformed data.
        """
        table_data = table_data.copy()
        combinations = table_data[self._columns].itertuples(index=False, name=None)
        uuids = map(self._combinations_to_uuids.get, combinations)
        table_data[self._joint_column] = list(uuids)
        return table_data.drop(self._columns, axis=1)

    def reverse_transform(self, table_data):
        """Reverse transform the table data.

        The transformation is reversed by popping the joint column from
        the table, mapping it back to the original combination of column values,
        and then setting all the columns back to the table with the original
        names.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Transformed data.
        """
        table_data = table_data.copy()
        columns = table_data.pop(self._joint_column).map(self._uuids_to_combinations)

        for index, column in enumerate(self._columns):
            table_data[column] = columns.str[index]

        return table_data


class GreaterThan(Constraint):
    """Ensure that the ``high`` column is always greater than the ``low`` one.

    The transformation strategy works by creating a column with the
    difference between the ``high`` and ``low`` values and then computing back the
    necessary columns using the difference and whichever other value is available.
    For example, if the ``high`` column is dropped, then the ``low`` column/value
    will be added to the diff to reconstruct the ``high`` column.

    Args:
        low (str or list[str]):
            Either the name of the column(s) that contains the low value,
            or a scalar that is the low value.
        high (str or list[str]):
            Either the name of the column(s) that contains the high value,
            or a scalar that is the high value.
        strict (bool):
            Whether the comparison of the values should be strict ``>=`` or
            not ``>`` when comparing them. Currently, this is only respected
            if ``reject_sampling`` or ``all`` handling strategies are used.
        handling_strategy (str):
            How this Constraint should be handled, which can be ``transform``
            or ``reject_sampling``. Defaults to ``transform``.
        fit_columns_model (bool):
            If False, reject sampling will be used to handle conditional sampling.
            Otherwise, a model will be trained and used to sample other columns
            based on the conditioned column. Defaults to False.
        drop (str):
            Which column to drop during transformation. Can be ``'high'``,
            ``'low'`` or ``None``.
        scalar (str):
            Which value is a scalar. Can be ``'high'``, ``'low'`` or ``None``.
            If ``None`` then both ``high`` and ``low`` are column names.
    """

    _diff_columns = None
    _is_datetime = None
    _columns_to_reconstruct = None

    @staticmethod
    def _as_list(value):
        if not isinstance(value, list):
            return [value]

        return value

    @staticmethod
    def _validate_scalar(scalar_column, column_names, scalar):
        """Validate scalar comparison inputs.

        - Make sure that the scalar column is not a list and raise the proper error if it is.
        - If the `column_names` is not a list it would make it a list.
        - Return both the scalar column and column names with the right format
        """
        if isinstance(scalar_column, list):
            raise TypeError(f'`{scalar}` cannot be a list when scalar="{scalar}".')

        column_names = GreaterThan._as_list(column_names)

        return column_names

    @staticmethod
    def _validate_drop(scalar, drop):
        if drop == scalar:
            raise ValueError(f"Invalid `drop` value: f`{drop}`. Cannot drop a scalar.")

    @classmethod
    def _validate_inputs(cls, low, high, scalar, drop):
        if scalar is None:
            low = cls._as_list(low)
            high = cls._as_list(high)
            if len(low) > 1 and len(high) > 1:
                raise ValueError('either `high` or `low` must contain only one column.')

            constraint_columns = tuple(low + high)

        elif scalar == 'low':
            cls._validate_drop(scalar, drop)
            high = cls._validate_scalar(scalar_column=low, column_names=high, scalar=scalar)
            constraint_columns = tuple(high)
            if isinstance(low, pd.Timestamp):
                low = low.to_datetime64()

        elif scalar == 'high':
            cls._validate_drop(scalar, drop)
            low = cls._validate_scalar(scalar_column=high, column_names=low, scalar=scalar)
            constraint_columns = tuple(low)
            if isinstance(high, pd.Timestamp):
                high = high.to_datetime64()

        else:
            raise ValueError(f"Invalad `scalar` value: `{scalar}`. "
                             "Use either: 'high', 'low', or None.")

        return low, high, constraint_columns

    def _get_columns_to_reconstruct(self):
        if self._drop == 'high':
            column = self._high
        elif self._drop == 'low':
            column = self._low
        elif self._scalar == 'high':
            column = self._low
        else:
            column = self._high

        return column

    def __init__(self, low, high, strict=False, handling_strategy='transform',
                 fit_columns_model=False, drop=None, scalar=None):
        self._strict = strict
        self._drop = drop
        self._scalar = scalar
        self._low, self._high, self.constraint_columns = self._validate_inputs(
            low=low, high=high, scalar=scalar, drop=drop)
        self._columns_to_reconstruct = self._get_columns_to_reconstruct()
        self.rebuild_columns = self._columns_to_reconstruct.copy()

        if strict:
            self.operator = np.greater
        else:
            self.operator = np.greater_equal

        super().__init__(handling_strategy=handling_strategy,
                         fit_columns_model=fit_columns_model)

    def _get_value(self, table_data, field):
        variable = getattr(self, f'_{field}')
        if self._scalar == field:
            return variable

        return table_data[variable].values

    def _get_diff_columns_name(self, table_data):
        names = []
        base = ''
        column_names = list(self.constraint_columns)
        if self._scalar is None:
            base = self._low if len(self._low) == 1 else self._high
            column_names.remove(base[0])
            base = str(base[0])

        for column in list(map(str, column_names)):
            token = '#'
            name = token.join((column, base))
            while name in table_data.columns:
                token += '#'

            names.append(name)

        return names

    def _get_is_datetime(self, table_data):
        low = self._get_value(table_data, 'low')
        high = self._get_value(table_data, 'high')

        is_low_datetime = is_datetime_type(low)
        is_high_datetime = is_datetime_type(high)
        is_datetime = is_low_datetime and is_high_datetime

        if not is_datetime and any([is_low_datetime, is_high_datetime]):
            raise ValueError('Both high and low must be datetime.')

        return is_datetime

    def _check_columns_exist(self, table_data, field):
        values = getattr(self, f'_{field}')
        missing = set(values) - set(table_data.columns)
        if missing:
            raise KeyError(f'The `{field}` columns {missing} '
                           f'were not found in table_data. If `{field}` is a scalar, '
                           f'set `scalar="{field}"`.')

    def _fit(self, table_data):
        """Learn the dtype of the high column.

        Args:
            table_data (pandas.DataFrame):
                The Table data.
        """
        if self._scalar != 'high':
            self._check_columns_exist(table_data, 'high')
        if self._scalar != 'low':
            self._check_columns_exist(table_data, 'low')

        self._dtype = table_data[self._columns_to_reconstruct].dtypes
        self._diff_columns = self._get_diff_columns_name(table_data)
        self._is_datetime = self._get_is_datetime(table_data)

    def is_valid(self, table_data):
        """Say whether ``high`` is greater than ``low`` in each row.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.Series:
                Whether each row is valid.
        """
        low = self._get_value(table_data, 'low')
        high = self._get_value(table_data, 'high')
        isnull = np.logical_or(np.isnan(low), np.isnan(high))

        valid = np.logical_or(self.operator(high, low), isnull)

        return valid.all(axis=1)

    def _transform(self, table_data):
        """Transform the table data.

        The transformation consist on replacing the ``high`` value with difference
        between it and the ``low`` value.

        Afterwards, a logarithm is applied to the difference + 1 to be able to ensure
        that the value stays positive when reverted afterwards using an exponential.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Transformed data.
        """
        table_data = table_data.copy()
        diff = self._get_value(table_data, 'high') - self._get_value(table_data, 'low')

        if self._is_datetime:
            diff = diff.astype(np.float64)

        table_data[self._diff_columns] = np.log(diff + 1)
        if self._drop == 'high':
            table_data = table_data.drop(self._high, axis=1)
        elif self._drop == 'low':
            table_data = table_data.drop(self._low, axis=1)

        return table_data

    def _construct_columns(self, diff, column_values, columns):
        new_values = pd.DataFrame(diff + column_values, columns=columns)
        return new_values.astype(dict(zip(columns, self._dtype)))

    def reverse_transform(self, table_data):
        """Reverse transform the table data.

        The transformation is reversed by computing an exponential of the given
        value, converting it to the original dtype, subtracting 1 and finally
        clipping the value to 0 on the low end to ensure the value is positive.

        Finally, the obtained value is added to the ``low`` column to get the final
        ``high`` value.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Transformed data.
        """
        table_data = table_data.copy()
        diff = (np.exp(table_data[self._diff_columns].values).round() - 1).clip(0)
        if self._is_datetime:
            diff = diff.astype('timedelta64[ns]')

        if self._drop == 'high':
            low = self._get_value(table_data, 'low')
            table_data[self._high] = self._construct_columns(diff, low, self._high)
        elif self._drop == 'low':
            high = self._get_value(table_data, 'high')
            table_data[self._low] = self._construct_columns(-diff, high, self._low)
        else:
            low = self._get_value(table_data, 'low')
            high = self._get_value(table_data, 'high')
            invalid = ~self.is_valid(table_data)
            if self._scalar == 'high':
                new_values = high - diff[invalid]
            elif self._scalar == 'low':
                new_values = low + diff[invalid]
            else:
                new_values = low[invalid] + diff[invalid]

            for i, column in enumerate(self._columns_to_reconstruct):
                table_data.loc[invalid, column] = new_values[:, i].astype(self._dtype[i])

        table_data = table_data.drop(self._diff_columns, axis=1)

        return table_data


class Positive(GreaterThan):
    """Ensure that the given column(s) are always positive.

    The transformation strategy works by creating columns with the
    difference between given columns and zero then computing back the
    necessary columns using the difference.

    Args:
        columns (str or list[str]):
            The name of the column(s) that are constrained to be positive.
        strict (bool):
            Whether the comparison of the values should be strict; disclude
            zero ``>`` or include it ``>=``. Currently, this is only respected
            if ``reject_sampling`` or ``all`` handling strategies are used.
        handling_strategy (str):
            How this Constraint should be handled, which can be ``transform``
            or ``reject_sampling``. Defaults to ``transform``.
        fit_columns_model (bool):
            If False, reject sampling will be used to handle conditional sampling.
            Otherwise, a model will be trained and used to sample other columns
            based on the conditioned column. Defaults to False.
        drop (bool):
            Whether to drop columns during transformation.
    """

    def __init__(self, columns, strict=False, handling_strategy='transform',
                 fit_columns_model=False, drop=False):
        drop = 'high' if drop else None
        super().__init__(handling_strategy=handling_strategy,
                         fit_columns_model=fit_columns_model,
                         high=columns, low=0, scalar='low',
                         drop=drop, strict=strict)


class Negative(GreaterThan):
    """Ensure that the given columns are always negative.

    The transformation strategy works by creating columns with the
    difference between zero and given columns then computing back the
    necessary columns using the difference.

    Args:
        columns (str or list[str]):
            The name of the column(s) that are constrained to be negative.
        strict (bool):
            Whether the comparison of the values should be strict, disclude
            zero ``<`` or include it ``<=``. Currently, this is only respected
            if ``reject_sampling`` or ``all`` handling strategies are used.
        handling_strategy (str):
            How this Constraint should be handled, which can be ``transform``
            or ``reject_sampling``. Defaults to ``transform``.
        fit_columns_model (bool):
            If False, reject sampling will be used to handle conditional sampling.
            Otherwise, a model will be trained and used to sample other columns
            based on the conditioned column. Defaults to False.
        drop (bool):
            Whether to drop columns during transformation.
    """

    def __init__(self, columns, strict=False, handling_strategy='transform',
                 fit_columns_model=False, drop=False):
        drop = 'low' if drop else None
        super().__init__(handling_strategy=handling_strategy,
                         fit_columns_model=fit_columns_model,
                         high=0, low=columns, scalar='high',
                         drop=drop, strict=strict)


class ColumnFormula(Constraint):
    """Compute a column based on applying a formula on the others.

    This contraint accepts as input a simple function and a column name.
    During the transformation phase the column is simply dropped.
    During the reverse transformation, the column is re-generated by
    applying the whole table to the given function.

    Args:
        column (str):
            Name of the column to compute applying the formula.
        formula (callable):
            Function to use for the computation.
        handling_strategy (str):
            How this Constraint should be handled, which can be ``transform``
            or ``reject_sampling``. Defaults to ``transform``.
        drop_column (bool):
            Whether or not to drop the constraint column.
    """

    def __init__(self, column, formula, handling_strategy='transform', drop_column=True):
        self._column = column
        self.constraint_columns = (column,)
        self._formula = import_object(formula)
        self._drop_column = drop_column
        self.rebuild_columns = (column,)
        super().__init__(handling_strategy, fit_columns_model=False)

    def is_valid(self, table_data):
        """Say whether the data fulfills the formula.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.Series:
                Whether each row is valid.
        """
        computed = self._formula(table_data)
        isnan = table_data[self._column].isna() & computed.isna()

        return table_data[self._column].eq(computed) | isnan

    def _transform(self, table_data):
        """Transform the table data.

        The transformation consist on simply dropping the indicated column from the
        table to prevent it from being modeled.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Transformed data.
        """
        table_data = table_data.copy()

        if self._drop_column and self._column in table_data:
            del table_data[self._column]

        return table_data

    def reverse_transform(self, table_data):
        """Reverse transform the table data.

        The transformation is reversed by applying the given formula function
        to the complete table and storing the result in the indicated column.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Transformed data.
        """
        table_data = table_data.copy()
        table_data[self._column] = self._formula(table_data)

        return table_data


class Between(Constraint):
    """Ensure that the ``constraint_column`` is always between ``high`` and ``low``.

    The transformation strategy works by replacing the ``constraint_column`` with a
    scaled version and then applying a logit function. The reverse transform
    applies a sigmoid to the data and then scales it back to the original space.

    Args:
        constraint_column (str):
            Name of the column to which the constraint will be applied.
        low (float or str):
            If float, lower bound on the values of the ``constraint_column``.
            If string, name of the column which will be the lower bound.
        high (float or str):
            If float, upper bound on the values of the ``constraint_column``.
            If string, name of the column which will be the upper bound.
        strict (bool):
            Whether the comparison of the values should be strict ``>=`` or
            not ``>`` when comparing them. Currently, this is only respected
            if ``reject_sampling`` or ``all`` handling strategies are used.
        handling_strategy (str):
            How this Constraint should be handled, which can be ``transform``
            or ``reject_sampling``. Defaults to ``transform``.
        fit_columns_model (bool):
            If False, reject sampling will be used to handle conditional sampling.
            Otherwise, a model will be trained and used to sample other columns
            based on the conditioned column. Defaults to False.
        high_is_scalar(bool or None):
            Whether or not the value for high is a scalar or a column name.
            If ``None``, this will be determined during the ``fit`` method
            by checking if the value provided is a column name.
        low_is_scalar(bool or None):
            Whether or not the value for low is a scalar or a column name.
            If ``None``, this will be determined during the ``fit`` method
            by checking if the value provided is a column name.
    """

    _transformed_column = None

    def __init__(self, column, low, high, strict=False, handling_strategy='transform',
                 fit_columns_model=False, high_is_scalar=None, low_is_scalar=None):
        self.constraint_column = column
        self.constraint_columns = (column,)
        self.rebuild_columns = (column,)
        self._strict = strict
        self._high_is_scalar = high_is_scalar
        self._low_is_scalar = low_is_scalar
        self._lt = operator.lt if strict else operator.le

        self._low = low
        if self._low_is_scalar and isinstance(low, pd.Timestamp):
            self._low = low.to_datetime64()

        self._high = high
        if self._high_is_scalar and isinstance(high, pd.Timestamp):
            self._high = high.to_datetime64()

        super().__init__(handling_strategy=handling_strategy,
                         fit_columns_model=fit_columns_model)

    def _get_low_value(self, table_data):
        """Return the appropriate lower bound.

        Returns the lower bound either as a column or a scalar, depending on the
        value of ``self._low_is_scalar``. If the lower bound column doesn't exist, returns
        ``None`` instead.

        Args:
            table_data (pandas.DataFrame):
                The Table data.

        Returns:
            pandas.DataFrame, float or None:
                Either the lower bound or None if the column doesn't exist.
        """
        if self._low_is_scalar:
            return self._low
        elif self._low in table_data.columns:
            return table_data[self._low]

        return None

    def _get_high_value(self, table_data):
        """Return the appropriate upper bound.

        Returns the upper bound either as a column or a scalar, depending on the
        value of ``self._high_is_scalar``. If the upper bound column doesn't exist, returns
        ``None`` instead.

        Args:
            table_data (pandas.DataFrame):
                The Table data.

        Returns:
            pandas.DataFrame, float or None:
                Either the upper bound or None if the column doesn't exist.
        """
        if self._high_is_scalar:
            return self._high
        elif self._high in table_data.columns:
            return table_data[self._high]

        return None

    def _get_diff_column_name(self, table_data):
        token = '#'
        components = list(map(str, [self.constraint_column, self._low, self._high]))
        while token.join(components) in table_data.columns:
            token += '#'

        return token.join(components)

    def _get_is_datetime(self, table_data):
        low = self._get_low_value(table_data)
        high = self._get_high_value(table_data)
        column = table_data[self.constraint_column]

        is_low_datetime = is_datetime_type(low)
        is_high_datetime = is_datetime_type(high)
        is_column_datetime = is_datetime_type(column)
        is_datetime = is_low_datetime and is_high_datetime and is_column_datetime

        if not is_datetime and any([is_low_datetime, is_high_datetime, is_column_datetime]):
            raise ValueError('The constraint column and bounds must all be datetime.')

        return is_datetime

    def _fit(self, table_data):
        if self._high_is_scalar is None:
            self._high_is_scalar = self._high not in table_data.columns
        if self._low_is_scalar is None:
            self._low_is_scalar = self._low not in table_data.columns

        self._transformed_column = self._get_diff_column_name(table_data)
        self._is_datetime = self._get_is_datetime(table_data)

    def is_valid(self, table_data):
        """Say whether the ``constraint_column`` is between the ``low`` and ``high`` values.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.Series:
                Whether each row is valid.
        """
        low = self._get_low_value(table_data)
        high = self._get_high_value(table_data)

        satisfy_low_bound = np.logical_or(
            self._lt(low, table_data[self.constraint_column]),
            np.isnan(low),
        )
        satisfy_high_bound = np.logical_or(
            self._lt(table_data[self.constraint_column], high),
            np.isnan(high),
        )

        return np.logical_or(
            np.logical_and(satisfy_low_bound, satisfy_high_bound),
            np.isnan(table_data[self.constraint_column]),
        )

    def _transform(self, table_data):
        """Transform the table data.

        The transformation consists of scaling the ``constraint_column``
        (``(column-low)/(high-low) * cnt + small_cnt``) and then applying
        a logit function to the scaled version of the column.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Transformed data.
        """
        table_data = table_data.copy()
        low = self._get_low_value(table_data)
        high = self._get_high_value(table_data)

        data = (table_data[self.constraint_column] - low) / (high - low)
        data = data * 0.95 + 0.025
        data = np.log(data / (1.0 - data))

        table_data[self._transformed_column] = data
        table_data = table_data.drop(self.constraint_column, axis=1)

        return table_data

    def reverse_transform(self, table_data):
        """Reverse transform the table data.

        The reverse transform consists of applying a sigmoid to the transformed
        ``constraint_column`` and then scaling it back to the original space
        ( ``(column - cnt) * (high - low) / cnt + low`` ).

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Transformed data.
        """
        table_data = table_data.copy()
        low = self._get_low_value(table_data)
        high = self._get_high_value(table_data)
        data = table_data[self._transformed_column]

        data = 1 / (1 + np.exp(-data))
        data = (data - 0.025) / 0.95
        data = data * (high - low) + low
        data = data.clip(low, high)

        if self._is_datetime:
            table_data[self.constraint_column] = pd.to_datetime(data)
        else:
            table_data[self.constraint_column] = data

        table_data = table_data.drop(self._transformed_column, axis=1)

        return table_data


class Rounding(Constraint):
    """Round a column based on the specified number of digits.

    Args:
        columns (str or list[str]):
            Name of the column(s) to round.
        digits (int):
            How much to round each column. All columns will be rounded to this
            number of digits.
        handling_strategy (str):
            How this Constraint should be handled, which can be ``transform``
            or ``reject_sampling``. Defaults to ``transform``.
        tolerance (int):
            When reject sampling, the sample data must be within this distance
            of the desired rounded values.
    """

    def __init__(self, columns, digits, handling_strategy='transform', tolerance=None):
        if digits > 15:
            raise ValueError('The value of digits cannot exceed 15')

        if tolerance is not None and tolerance >= 10**(-1 * digits):
            raise ValueError('Tolerance must be less than the rounding level')

        if isinstance(columns, str):
            self._columns = [columns]
        else:
            self._columns = columns

        self._digits = digits
        self._round_config = {column: self._digits for column in self._columns}
        self._tolerance = tolerance if tolerance else 10**(-1 * (digits + 1))
        super().__init__(handling_strategy=handling_strategy, fit_columns_model=False)

    def is_valid(self, table_data):
        """Determine if the data satisfies the rounding constraint.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.Series:
                Whether each row is valid.
        """
        columns = table_data[self._columns]
        rounded = columns.round(self._digits)
        valid = (columns - rounded).abs() <= self._tolerance

        return valid.all(1)

    def reverse_transform(self, table_data):
        """Reverse transform the table data.

        Round the columns to the desired digits.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Transformed data.
        """
        return table_data.round(self._round_config)


class OneHotEncoding(Constraint):
    """Ensure the appropriate columns are one hot encoded.

    This constraint allows the user to specify a list of columns where each row
    is a one hot vector. During the reverse transform, the output of the model
    is transformed so that the column with the largest value is set to 1 while
    all other columns are set to 0.

    Args:
        columns (list[str]):
            Names of the columns containing one hot rows.
        handling_strategy (str):
            How this Constraint should be handled, which can be ``transform``
            or ``reject_sampling`` (not recommended). Defaults to ``transform``.
    """

    def __init__(self, columns, handling_strategy='transform'):
        self._columns = columns
        self.constraint_columns = tuple(columns)
        super().__init__(handling_strategy, fit_columns_model=True)

    def _sample_constraint_columns(self, table_data):
        """Handle constraint columns when conditioning.

        When handling a set of one-hot columns, a subset of columns may be provided
        to condition on. To handle this, this function does the following:

        1. If the user specifies that a particular column must be 1,
           then all other columns must be 0.
        2. If the user specifies that one or more columns must be 0, then
           we need to sample the other columns and select the highest value
           and enforce the one-hot constraint.
        3. If the user specifies something invalid, we need to raise an error.

        Args:
            table_data (pandas.DataFrame):
                Table data containing the conditions.

        Returns:
            pandas.DataFrame:
                Table data with the constraint columns filled in.

        Raise:
            ``ValueError`` if the conditions are invalid.
        """
        table_data = table_data.copy()

        condition_columns = [col for col in self._columns if col in table_data.columns]
        conditions_data = table_data[condition_columns]
        conditions_data_sum = conditions_data.sum(axis=1)
        if not conditions_data.isin([0.0, 1.0]).all(axis=1).all():
            raise ValueError('Condition values must be ones or zeros.')

        if (conditions_data_sum > 1.0).any():
            raise ValueError('Each row of a condition can only contain one number one.')

        has_one = conditions_data_sum == 1.0
        if (~has_one).sum() > 0:
            sub_table_data = table_data.loc[~has_one, condition_columns]
            should_transform = False

            if len(condition_columns) == len(self._columns) - 1:
                proposed_table_data = sub_table_data.copy()
                for column in self._columns:
                    if column not in condition_columns:
                        proposed_table_data[column] = 1.0

            else:
                should_transform = True
                conditions = sub_table_data[condition_columns]
                transformed_conditions = self._hyper_transformer.transform(conditions)
                proposed_table_data = self._columns_model.sample(
                    num_rows=len(sub_table_data),
                    conditions=transformed_conditions.iloc[0].to_dict()
                )

            if should_transform:
                proposed_table_data = self._hyper_transformer.reverse_transform(
                    proposed_table_data)

            for column in self._columns:
                if column not in condition_columns:
                    sub_table_data[column] = proposed_table_data[column].values
                else:
                    sub_table_data[column] = float('-inf')

            table_data.loc[~has_one, self._columns] = self.reverse_transform(sub_table_data)

        if has_one.sum() > 0:
            for column in self._columns:
                if column not in condition_columns:
                    table_data.loc[has_one, column] = 0

        return table_data

    def is_valid(self, table_data):
        """Check whether the data satisfies the one-hot constraint.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.Series:
                Whether each row is valid.
        """
        one_hot_data = table_data[self._columns]

        sum_one = one_hot_data.sum(axis=1) == 1.0
        max_one = one_hot_data.max(axis=1) == 1.0
        min_zero = one_hot_data.min(axis=1) == 0.0

        return sum_one & max_one & min_zero

    def reverse_transform(self, table_data):
        """Reverse transform the table data.

        Set the column with the largest value to one, set all other columns to zero.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Transformed data.
        """
        table_data = table_data.copy()

        one_hot_data = table_data[self._columns]
        transformed_data = np.zeros_like(one_hot_data.values)
        transformed_data[np.arange(len(one_hot_data)), np.argmax(one_hot_data.values, axis=1)] = 1
        table_data[self._columns] = transformed_data

        return table_data


class Unique(Constraint):
    """Ensure that each value for a specified column/group of columns is unique.

    This constraint is provided a list of columns, and guarantees that every
    unique combination of those columns appears at most once in the sampled
    data.

    Args:
        columns (str or list[str]):
            Name of the column(s) to keep unique.
    """

    def __init__(self, columns):
        self.columns = columns if isinstance(columns, list) else [columns]
        super().__init__(handling_strategy='reject_sampling', fit_columns_model=False)

    def is_valid(self, table_data):
        """Get indices of first instance of unique rows.

        If a row is the first instance of that combination of column
        values, it is valid. Otherwise it is false.

        Args:
            table_data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.Series:
                Whether each row is valid.
        """
        return table_data.groupby(self.columns).cumcount() == 0



from typing import Union

import numpy as np
import pandas as pd
from sdmetrics.utils import is_datetime


def get_histograms(
    original: pd.Series,
    synthetic: pd.Series,
    normalize: bool = True,
    bins: Union[str, np.array] = "doane",
    return_keys: bool = False,
) -> tuple:
    """Get percentual frequencies or counts for each possible real categorical value.

    Returns:
        The observed and expected frequencies.
    """

    if is_datetime(original):
        original = pd.to_numeric(original, errors="coerce", downcast="integer")
        synthetic = pd.to_numeric(synthetic, errors="coerce", downcast="integer")

    if original.dtype.name in ("object", "category", "bool"):  # categorical
        gt = original.value_counts().to_dict()
        synth = synthetic.value_counts().to_dict()
        all_keys = gt.keys() | synth.keys()
        for key in all_keys:
            gt.setdefault(key, 0)
            synth.setdefault(key, 0)
    elif np.issubdtype(original.dtype, np.number):  # continuous
        original = original.dropna()
        synthetic = synthetic.dropna()
        if type(bins) is not np.ndarray:
            combined = pd.concat([original, synthetic])
            bins = np.histogram_bin_edges(combined, bins=bins)
        gt_vals, _ = np.histogram(original, bins=bins)
        synth_vals, _ = np.histogram(synthetic, bins=bins)
        gt = {k: v for k, v in zip(bins, gt_vals)}
        synth = {k: v for k, v in zip(bins, synth_vals)}
    else:
        raise ValueError("Column is not categorical or continouous")

    # order the keys
    gt = {k: v for k, v in sorted(gt.items(), key=lambda item: item[0])}
    synth = {k: v for k, v in sorted(synth.items(), key=lambda item: item[0])}

    assert gt.keys() == synth.keys(), "Keys do not match for column"

    if normalize:
        gt_sum = sum(gt.values())
        synth_sum = sum(synth.values())
        gt = {k: v / gt_sum for k, v in gt.items()}
        synth = {k: v / synth_sum for k, v in synth.items()}

    frequencies = (np.array(list(gt.values())), np.array(list(synth.values())))
    if return_keys:
        return frequencies, list(gt.keys())

    return frequencies



"""Wrapper around CTGAN model."""

import numpy as np
from rctgan.ctganpc import PC_CTGANSynthesizer, CTGANSynthesizer
import logging
from rctgan.metadata import Table
from rctgan.tabular.base import BaseTabularModel
import numpy as np
import pandas as pd
import math
import os
import uuid

LOGGER = logging.getLogger(__name__)
COND_IDX = str(uuid.uuid4())
FIXED_RNG_SEED = 73251
TMP_FILE_NAME = '.sample.csv.temp'
DISABLE_TMP_FILE = 'disable'

class CTGANModel(BaseTabularModel):
    """Base class for all the CTGAN models.

    The ``CTGANModel`` class provides a wrapper for all the CTGAN models.
    """

    _MODEL_CLASS = None
    _model_kwargs = None

    _DTYPE_TRANSFORMERS = {
        'O': None
    }

    def _build_model(self):
        return self._MODEL_CLASS(**self._model_kwargs)

    def _fit(self, table_data):
        """Fit the model to the table.

        Args:
            table_data (pandas.DataFrame):
                Data to be learned.
        """
        self._model = self._build_model()

        categoricals = []
        fields_before_transform = self._metadata.get_fields()
        for field in table_data.columns:
            if field in fields_before_transform:
                meta = fields_before_transform[field]
                if meta['type'] == 'categorical':
                    categoricals.append(field)

            else:
                field_data = table_data[field].dropna()
                if set(field_data.unique()) == {0.0, 1.0}:
                    # booleans encoded as float values must be modeled as bool
                    field_data = field_data.astype(bool)

                dtype = field_data.infer_objects().dtype
                try:
                    kind = np.dtype(dtype).kind
                except TypeError:
                    # probably category
                    kind = 'O'
                if kind in ['O', 'b']:
                    categoricals.append(field)
        self._model.fit(
            table_data,
            discrete_columns=categoricals
        )

    def _sample(self, num_rows, conditions=None):
        """Sample the indicated number of rows from the model.

        Args:
            num_rows (int):
                Amount of rows to sample.
            conditions (dict):
                If specified, this dictionary maps column names to the column
                value. Then, this method generates `num_rows` samples, all of
                which are conditioned on the given variables.

        Returns:
            pandas.DataFrame:
                Sampled data.
        """
        if conditions is None:
            return self._model.sample(num_rows)

        raise NotImplementedError(f"{self._MODEL_CLASS} doesn't support conditional sampling.")

    def _set_random_state(self, random_state):
        """Set the random state of the model's random number generator.

        Args:
            random_state (int, tuple[np.random.RandomState, torch.Generator], or None):
                Seed or tuple of random states to use.
        """
        self._model.set_random_state(random_state)


class CTGAN(CTGANModel):
    """Model wrapping ``CTGANSynthesizer`` model.

    Args:
        field_names (list[str]):
            List of names of the fields that need to be modeled
            and included in the generated output data. Any additional
            fields found in the data will be ignored and will not be
            included in the generated output.
            If ``None``, all the fields found in the data are used.
        field_types (dict[str, dict]):
            Dictinary specifying the data types and subtypes
            of the fields that will be modeled. Field types and subtypes
            combinations must be compatible with the SDV Metadata Schema.
        field_transformers (dict[str, str]):
            Dictinary specifying which transformers to use for each field.
            Available transformers are:

                * ``integer``: Uses a ``NumericalTransformer`` of dtype ``int``.
                * ``float``: Uses a ``NumericalTransformer`` of dtype ``float``.
                * ``categorical``: Uses a ``CategoricalTransformer`` without gaussian noise.
                * ``categorical_fuzzy``: Uses a ``CategoricalTransformer`` adding gaussian noise.
                * ``one_hot_encoding``: Uses a ``OneHotEncodingTransformer``.
                * ``label_encoding``: Uses a ``LabelEncodingTransformer``.
                * ``boolean``: Uses a ``BooleanTransformer``.
                * ``datetime``: Uses a ``DatetimeTransformer``.

        anonymize_fields (dict[str, str]):
            Dict specifying which fields to anonymize and what faker
            category they belong to.
        primary_key (str):
            Name of the field which is the primary key of the table.
        constraints (list[Constraint, dict]):
            List of Constraint objects or dicts.
        table_metadata (dict or metadata.Table):
            Table metadata instance or dict representation.
            If given alongside any other metadata-related arguments, an
            exception will be raised.
            If not given at all, it will be built using the other
            arguments or learned from the data.
        embedding_dim (int):
            Size of the random sample passed to the Generator. Defaults to 128.
        generator_dim (tuple or list of ints):
            Size of the output samples for each one of the Residuals. A Residual Layer
            will be created for each one of the values provided. Defaults to (256, 256).
        discriminator_dim (tuple or list of ints):
            Size of the output samples for each one of the Discriminator Layers. A Linear Layer
            will be created for each one of the values provided. Defaults to (256, 256).
        generator_lr (float):
            Learning rate for the generator. Defaults to 2e-4.
        generator_decay (float):
            Generator weight decay for the Adam Optimizer. Defaults to 1e-6.
        discriminator_lr (float):
            Learning rate for the discriminator. Defaults to 2e-4.
        discriminator_decay (float):
            Discriminator weight decay for the Adam Optimizer. Defaults to 1e-6.
        batch_size (int):
            Number of data samples to process in each step.
        discriminator_steps (int):
            Number of discriminator updates to do for each generator update.
            From the WGAN paper: https://arxiv.org/abs/1701.07875. WGAN paper
            default is 5. Default used is 1 to match original CTGAN implementation.
        log_frequency (boolean):
            Whether to use log frequency of categorical levels in conditional
            sampling. Defaults to ``True``.
        verbose (boolean):
            Whether to have print statements for progress results. Defaults to ``False``.
        epochs (int):
            Number of training epochs. Defaults to 300.
        pac (int):
            Number of samples to group together when applying the discriminator.
            Defaults to 10.
        cuda (bool or str):
            If ``True``, use CUDA. If a ``str``, use the indicated device.
            If ``False``, do not use cuda at all.
        rounding (int, str or None):
            Define rounding scheme for ``NumericalTransformer``. If set to an int, values
            will be rounded to that number of decimal places. If ``None``, values will not
            be rounded. If set to ``'auto'``, the transformer will round to the maximum number
            of decimal places detected in the fitted data. Defaults to ``'auto'``.
        min_value (int, str or None):
            Specify the minimum value the ``NumericalTransformer`` should use. If an integer
            is given, sampled data will be greater than or equal to it. If the string ``'auto'``
            is given, the minimum will be the minimum value seen in the fitted data. If ``None``
            is given, there won't be a minimum. Defaults to ``'auto'``.
        max_value (int, str or None):
            Specify the maximum value the ``NumericalTransformer`` should use. If an integer
            is given, sampled data will be less than or equal to it. If the string ``'auto'``
            is given, the maximum will be the maximum value seen in the fitted data. If ``None``
            is given, there won't be a maximum. Defaults to ``'auto'``.
    """

    _MODEL_CLASS = CTGANSynthesizer

    def __init__(self, field_names=None, field_types=None, field_transformers=None,
                 anonymize_fields=None, primary_key=None, constraints=None, table_metadata=None,
                 embedding_dim=128, generator_dim=(256, 256), discriminator_dim=(256, 256),
                 generator_lr=2e-4, generator_decay=1e-6, discriminator_lr=2e-4,
                 discriminator_decay=1e-6, batch_size=500, discriminator_steps=1,
                 log_frequency=True, verbose=False, epochs=300, pac=10, cuda=True, plot_loss=False, seed=None,
                 rounding='auto', min_value='auto', max_value='auto'):
        super().__init__(
            field_names=field_names,
            primary_key=primary_key,
            field_types=field_types,
            field_transformers=field_transformers,
            anonymize_fields=anonymize_fields,
            constraints=constraints,
            table_metadata=table_metadata,
            rounding=rounding,
            max_value=max_value,
            min_value=min_value
        )

        self._model_kwargs = {
            'embedding_dim': embedding_dim,
            'generator_dim': generator_dim,
            'discriminator_dim': discriminator_dim,
            'generator_lr': generator_lr,
            'generator_decay': generator_decay,
            'discriminator_lr': discriminator_lr,
            'discriminator_decay': discriminator_decay,
            'batch_size': batch_size,
            'discriminator_steps': discriminator_steps,
            'log_frequency': log_frequency,
            'verbose': verbose,
            'epochs': epochs,
            'pac': pac,
            'cuda': cuda,
            'plot_loss': plot_loss,
            'seed': seed
        }
        

class PC_CTGANModel(BaseTabularModel):
    """Base class for all the CTGAN models.

    The ``CTGANModel`` class provides a wrapper for all the CTGAN models.
    """

    _MODEL_CLASS = None
    _model_kwargs = None

    _DTYPE_TRANSFORMERS = {
        'O': None
    }
    
    _metadata_pc = None
    
    _type_dict = {}
    _model = None

    def _build_model(self):
        return self._MODEL_CLASS(**self._model_kwargs)

    def _fit(self, table_data, parent_data):
        """Fit the model to the table.

        Args:
            table_data (pandas.DataFrame):
                Data to be learned.
        """
        self._model = self._build_model()

        categoricals = []
        fields_before_transform = self._metadata.get_fields()
        for field in table_data.columns:
            if field in fields_before_transform:
                meta = fields_before_transform[field]
                if meta['type'] == 'categorical':
                    categoricals.append(field)

            else:
                field_data = table_data[field].dropna()
                if set(field_data.unique()) == {0.0, 1.0}:
                    # booleans encoded as float values must be modeled as bool
                    field_data = field_data.astype(bool)

                dtype = field_data.infer_objects().dtype
                try:
                    kind = np.dtype(dtype).kind
                except TypeError:
                    # probably category
                    kind = 'O'
                if kind in ['O', 'b']:
                    categoricals.append(field)
        self._type_dict = dict(table_data.dtypes)
        
        self._model.fit(
            table_data,
            parent_data,
            discrete_columns=categoricals
        )
        
    def fit(self, data, parent_data, field_names_pc=None, field_types_pc=None, field_transformers_pc=None,
                 anonymize_fields_pc=None, primary_key_pc=None, constraints_pc=None, table_metadata_pc=None,
                 rounding_pc='auto', min_value_pc='auto', max_value_pc='auto'):
        if isinstance(data, pd.DataFrame):
            data = data.reset_index(drop=True)

        LOGGER.debug('Fitting %s to table %s; shape: %s', self.__class__.__name__,
                     self._metadata.name, data.shape)
        if not self._metadata_fitted:
            self._metadata.fit(data)

        self._num_rows = len(data)

        LOGGER.debug('Transforming table %s; shape: %s', self._metadata.name, data.shape)
        transformed = self._metadata.transform(data)
        
        
        self._metadata_pc = Table(
                field_names=field_names_pc,
                primary_key=primary_key_pc,
                field_types=field_types_pc,
                field_transformers=field_transformers_pc,
                anonymize_fields=anonymize_fields_pc,
                constraints=constraints_pc,
                dtype_transformers=self._DTYPE_TRANSFORMERS,
                rounding=rounding_pc,
                min_value=min_value_pc,
                max_value=max_value_pc
            )
        self._metadata_pc.fit(parent_data)
        if isinstance(parent_data, pd.DataFrame):
            parent_data = parent_data.reset_index(drop=True)

        LOGGER.debug('Fitting %s to table %s; shape: %s', self.__class__.__name__,
                     self._metadata_pc.name, parent_data.shape)
        self._num_rows_pc = len(parent_data)

        LOGGER.debug('Transforming table %s; shape: %s', self._metadata_pc.name, parent_data.shape)
        transformed_pc = self._metadata_pc.transform(parent_data)

        if self._metadata.get_dtypes(ids=False) and self._metadata_pc.get_dtypes(ids=False):
            LOGGER.debug(
                'Fitting %s model to table %s', self.__class__.__name__, self._metadata.name)
            LOGGER.debug(
                'Fitting %s model to table %s', self.__class__.__name__, self._metadata_pc.name)
            self._fit(transformed, transformed_pc)
    
    def _randomize_samples(self, randomize_samples):
        """Randomize the samples according to user input.

        If ``randomize_samples`` is false, fix the seed that the random number generator
        uses in the underlying models.

        Args:
            randomize_samples (bool):
                Whether or not to randomize the generated samples.
        """
        if self._model is None:
            return

        if randomize_samples:
            self._set_random_state(None)
        else:
            self._set_random_state(FIXED_RNG_SEED)
    
    def _validate_file_path(self, output_file_path):
        """Validate the user-passed output file arg, and create the file."""
        output_path = None
        if output_file_path == DISABLE_TMP_FILE:
            # Temporary way of disabling the output file feature, used by HMA1.
            return output_path

        elif output_file_path:
            output_path = os.path.abspath(output_file_path)
            if os.path.exists(output_path):
                raise AssertionError(f'{output_path} already exists.')

        else:
            if os.path.exists(TMP_FILE_NAME):
                os.remove(TMP_FILE_NAME)

            output_path = TMP_FILE_NAME

        # Create the file.
        with open(output_path, 'w+'):
            pass

        return output_path
    
    def _sample(self, sizes, parent, conditions=None):
        if conditions is None:
            if self._model is None:
                def dupli_rows(data, size_list):
                    df = data.copy()
                    df['index_prim_key'] = range(len(df))
                    size_list_2 = []
                    for k in range(len(size_list)):
                        s = size_list[k]
                        size_list_2 += [k]*s
                    index_prim_key = pd.DataFrame(size_list_2, columns=['index_prim_key'])
                    df = index_prim_key.merge(df, on=['index_prim_key'], how='left')
                    df = df.drop(['index_prim_key'], axis=1)
                    return df
                data = pd.DataFrame(np.zeros((np.sum(sizes), 0)))
                if sum(sizes)==len(sizes):
                    data["Parent_index"] = range(len(data))
                else:
                    data["Parent_index"] = dupli_rows(pd.DataFrame(parent.index, columns=['__index__']), sizes)['__index__']
                return data
            return self._model.sample(sizes, parent)

        raise NotImplementedError(f"{self._MODEL_CLASS} doesn't support conditional sampling.")
    
    
    def sample(self, sizes, parent, randomize_samples=True, output_file_path=None,
               conditions=None):
        if conditions is not None:
            raise TypeError('This method does not support the conditions parameter. '
                            'Please create `sdvrctgan.sampling.Condition` objects and pass them '
                            'into the `sample_conditions` method. '
                            'See User Guide or API for more details.')

        if sizes is None:
            raise ValueError('You must specify the number of rows to sample (e.g. num_rows=100).')

        no_size = True
        for s in sizes:
            if s>0:
                no_size = False
                break
        if no_size or len(sizes)==0:
            return pd.DataFrame()

        self._randomize_samples(randomize_samples)

        output_file_path = self._validate_file_path(output_file_path)
        
        LOGGER.debug('Fitting %s to table %s; shape: %s', self.__class__.__name__,
                     self._metadata_pc.name, parent.shape)
        self._num_rows_pc = len(parent)

        LOGGER.debug('Transforming table %s; shape: %s', self._metadata_pc.name, parent.shape)
        _transformed_pc = self._metadata_pc.transform(parent)

        sampled = self._sample(
            sizes,
            _transformed_pc
        )

        parent_index = sampled["Parent_index"].copy()
        sampled.drop("Parent_index", inplace=True, axis=1)
        sampled = sampled.astype(self._type_dict)
        sampled = self._metadata.reverse_transform(sampled)
        sampled["Parent_index"] = parent_index
        
        return sampled
    
    
    
    def _set_random_state(self, random_state):
        """Set the random state of the model's random number generator.

        Args:
            random_state (int, tuple[np.random.RandomState, torch.Generator], or None):
                Seed or tuple of random states to use.
        """
        self._model.set_random_state(random_state)


class PC_CTGAN(PC_CTGANModel):

    _MODEL_CLASS = PC_CTGANSynthesizer

    def __init__(self, field_names=None, field_types=None, field_transformers=None,
                 anonymize_fields=None, primary_key=None, constraints=None, table_metadata=None,
                 embedding_dim=128, generator_dim=(256, 256), discriminator_dim=(256, 256),
                 generator_lr=2e-4, generator_decay=1e-6, discriminator_lr=2e-4,
                 discriminator_decay=1e-6, batch_size=500, discriminator_steps=1,
                 log_frequency=True, verbose=False, epochs=300, pac=10, cuda=True, plot_loss=False, seed=None,
                 rounding='auto', min_value='auto', max_value='auto'):
        super().__init__(
            field_names=field_names,
            primary_key=primary_key,
            field_types=field_types,
            field_transformers=field_transformers,
            anonymize_fields=anonymize_fields,
            constraints=constraints,
            table_metadata=table_metadata,
            rounding=rounding,
            max_value=max_value,
            min_value=min_value
        )

        self._model_kwargs = {
            'embedding_dim': embedding_dim,
            'generator_dim': generator_dim,
            'discriminator_dim': discriminator_dim,
            'generator_lr': generator_lr,
            'generator_decay': generator_decay,
            'discriminator_lr': discriminator_lr,
            'discriminator_decay': discriminator_decay,
            'batch_size': batch_size,
            'discriminator_steps': discriminator_steps,
            'log_frequency': log_frequency,
            'verbose': verbose,
            'epochs': epochs,
            'pac': pac,
            'cuda': cuda,
            'plot_loss': plot_loss,
            'seed': seed
        }



import argparse
from pathlib import Path

import pandas as pd
from sklearn.impute import SimpleImputer
from synthcity.plugins import Plugins
from syntherela.metadata import Metadata
from syntherela.data import load_tables, save_tables, remove_sdv_columns

Plugins(categories=["generic", "privacy"]).list()

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="airbnb-simplified_subsampled")
args.add_argument("--real-data-path", type=str, default="data/original")
args.add_argument("--synthetic-data-path", type=str, default="data/synthetic")
args.add_argument("--model-save-path", type=str, default="checkpoints")
args.add_argument("--run-id", type=str, default="1")
args = args.parse_args()
dataset_name = args.dataset_name
real_data_path = args.real_data_path
synthetic_data_path = args.synthetic_data_path
run_id = args.run_id

MODEL_NAMES = [
    "bayesian_network",
    "ddpm",
    "ctgan",
    "tvae",
    "nflow",
]

SYNTHETIC_MODEL_PARAMS = {
    "marginal_distributions": {},
    "tvae": {
        "n_iter": 300,
        "lr": 0.0002,
        "decoder_n_layers_hidden": 4,
        "weight_decay": 0.001,
        "batch_size": 256,
        "n_units_embedding": 200,
        "decoder_n_units_hidden": 300,
        "decoder_nonlin": "elu",
        "decoder_dropout": 0.194325119117226,
        "encoder_n_layers_hidden": 1,
        "encoder_n_units_hidden": 450,
        "encoder_nonlin": "leaky_relu",
        "encoder_dropout": 0.04288563703094718,
    },
    "ctgan": {
        "generator_n_layers_hidden": 2,
        "generator_n_units_hidden": 50,
        "generator_nonlin": "tanh",
        "n_iter": 1000,
        "generator_dropout": 0.0574657940165757,
        "discriminator_n_layers_hidden": 4,
        "discriminator_n_units_hidden": 150,
        "discriminator_nonlin": "relu",
        "discriminator_n_iter": 3,
        "discriminator_dropout": 0.08727454632095322,
        "lr": 0.0001,
        "weight_decay": 0.0001,
        "batch_size": 500,
        "encoder_max_clusters": 14,
    },
    "bayesian_network": {
        "struct_learning_search_method": "hillclimb",
        "struct_learning_score": "bic",
    },
    "nflow": {
        "n_iter": 1000,
        "n_layers_hidden": 10,
        "n_units_hidden": 98,
        "dropout": 0.11496088236749386,
        "batch_norm": True,
        "lr": 0.0001,
        "linear_transform_type": "permutation",
        "base_transform_type": "rq-autoregressive",
        "batch_size": 512,
    },
    "ddpm": {
        "lr": 0.0009375080542687667,
        "batch_size": 2929,
        "num_timesteps": 998,
        "n_iter": 1051,
        "is_classification": False,
    },
}

metadata = Metadata().load_from_json(
    Path(real_data_path) / f"{dataset_name}/metadata.json"
)
real_data = load_tables(Path(real_data_path) / f"{dataset_name}", metadata)
real_data, metadata = remove_sdv_columns(real_data, metadata, validate=False)
metadata.validate_data(real_data)

for MODEL_NAME in MODEL_NAMES:
    synthetic_data = {}
    for table in real_data.keys():
        print(f"{MODEL_NAME} - {table}")
        try:
            X = real_data[table]
            X_orig = X.copy()
            syn_model = Plugins().get(MODEL_NAME, **SYNTHETIC_MODEL_PARAMS[MODEL_NAME])
            syn_model.strict = False

            # check if any column is constant
            constant_columns = X.columns[X.nunique() == 1]
            if len(constant_columns) > 0:
                X = X.drop(columns=constant_columns)

            numeric_columns = X.select_dtypes(include="number").columns
            if len(numeric_columns) > 0:
                imp = SimpleImputer(strategy="mean")
                X[numeric_columns] = pd.DataFrame(
                    imp.fit_transform(X[numeric_columns]), columns=numeric_columns
                )

            syn_model.fit(X)
            synthetic_data[table] = syn_model.generate(len(X))
            synthetic_data[table] = synthetic_data[table].dataframe()

            # append comstant columns
            for column in constant_columns:
                synthetic_data[table][column] = X_orig[column]
        except Exception as e:
            print(f"Exception occured at {MODEL_NAME}-{table}: {e}")

    save_data_path = (
        Path(synthetic_data_path) / dataset_name / MODEL_NAME / run_id / "sample1"
    )
    save_tables(synthetic_data, save_data_path)



import sys
import logging
import argparse

from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from xgboost import XGBClassifier

from syntherela.benchmark import Benchmark
from syntherela.metrics.single_column.distance import (
    HellingerDistance,
    JensenShannonDistance,
    WassersteinDistance,
    TotalVariationDistance,
)
from syntherela.metrics.single_column.statistical import (
    ChiSquareTest,
    KolmogorovSmirnovTest,
)
from syntherela.metrics.single_table.distance import (
    MaximumMeanDiscrepancy,
    PairwiseCorrelationDifference,
)
from syntherela.metrics.single_column.detection import SingleColumnDetection
from syntherela.metrics.single_table.detection import SingleTableDetection

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="airbnb-simplified_subsampled")
args.add_argument("--methods", "-m", action="append", default=None)
args.add_argument("--run-id", type=str, default="1")
args = args.parse_args()
dataset_name = args.dataset_name
methods = args.methods
run_id = args.run_id

logger = logging.getLogger(f"{dataset_name}_logger")

logger.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(formatter)
logger.addHandler(handler)

logger.info(f"START LOGGING Dataset: {dataset_name}")

xgb_cls = XGBClassifier
xgb_args = {"seed": 0}
rf_cls = RandomForestClassifier
rf_args = {"random_state": 0, "n_estimators": 100}
logistic = LogisticRegression
logistic_args = {"random_state": 0}
single_column_metrics = [
    ChiSquareTest(),
    KolmogorovSmirnovTest(),
    TotalVariationDistance(),
    HellingerDistance(),
    JensenShannonDistance(),
    WassersteinDistance(),
    SingleColumnDetection(
        classifier_cls=xgb_cls, classifier_args=xgb_args, random_state=42
    ),
    # SingleColumnDetection(classifier_cls=rf_cls, classifier_args=rf_args),
    SingleColumnDetection(
        classifier_cls=logistic, classifier_args=logistic_args, random_state=42
    ),
]
single_table_metrics = [
    MaximumMeanDiscrepancy(),
    PairwiseCorrelationDifference(),
    SingleTableDetection(
        classifier_cls=xgb_cls, classifier_args=xgb_args, random_state=42
    ),
    # SingleTableDetection(classifier_cls=rf_cls, classifier_args=rf_args),
    SingleTableDetection(
        classifier_cls=logistic, classifier_args=logistic_args, random_state=42
    ),
]

benchmark = Benchmark(
    real_data_dir="data/original",
    synthetic_data_dir="data/synthetic",
    results_dir=f"results/{run_id}",
    benchmark_name="Benchmark",
    single_column_metrics=single_column_metrics,
    single_table_metrics=single_table_metrics,
    run_id=run_id,
    sample_id="sample1",
    datasets=[dataset_name],
    methods=methods,
    validate_metadata=False,
)

benchmark.run()



"""Constraint utility functions."""

from datetime import datetime

import pandas as pd


def is_datetime_type(val):
    """Determine if the input is a datetime type or not.

    Args:
        val (pandas.DataFrame, int or datetime):
            Input to evaluate.

    Returns:
        bool:
            True if the input is a datetime type, False if not.
    """
    return (
        pd.api.types.is_datetime64_any_dtype(val)
        or isinstance(val, pd.Timestamp)
        or isinstance(val, datetime)
    )



import numpy as np
import pandas as pd
from scipy.spatial.distance import jensenshannon
from sdmetrics.goal import Goal
from sdmetrics.utils import is_datetime

from syntherela.metrics.base import SingleColumnMetric, DistanceBaseMetric
from syntherela.metrics.single_column.distance.utils import get_histograms


class JensenShannonDistance(DistanceBaseMetric, SingleColumnMetric):
    def __init__(self, base=np.e, **kwargs):
        super().__init__(**kwargs)
        self.name = "JensenShannonDistance"
        self.goal = Goal.MINIMIZE
        self.base = base
        self.min_value = 0.0
        self.max_value = np.emath.logn(base, 2)

    @staticmethod
    def is_applicable(column_type):
        return column_type in ["categorical", "numerical", "datetime", "boolean"]

    @staticmethod
    def compute(
        orig_col, synth_col, bins, normalize_histograms=True, base=np.e, **kwargs
    ):
        gt_freq, synth_freq = get_histograms(
            orig_col, synth_col, normalize=normalize_histograms, bins=bins
        )
        return jensenshannon(gt_freq, synth_freq, base=base)

    def run(self, real_data, synthetic_data, **kwargs):
        if self.is_constant(real_data):
            return {
                "value": 0,
                "reference_ci": [0, 0],
                "bootstrap_mean": 0,
                "bootstrap_se": 0,
            }
        # check for datetime
        if is_datetime(real_data):
            real_data = pd.to_numeric(real_data, errors="coerce", downcast="integer")
            synthetic_data = pd.to_numeric(
                synthetic_data, errors="coerce", downcast="integer"
            )
        # compute bin values on the original data
        if real_data.dtype.name in ("object", "category", "bool"):
            bins = None
        else:
            real_data = real_data.dropna()
            synthetic_data = synthetic_data.dropna()
            bins = np.histogram_bin_edges(real_data)
        return super().run(
            real_data, synthetic_data, bins=bins, base=self.base, **kwargs
        )



import re
import json
import warnings

import pandas as pd
import numpy as np

warnings.filterwarnings("ignore")

run = 1
method_names_dict = {
    "SDV": "SDV",
    "RCTGAN": "RCTGAN",
    "REALTABFORMER": "REALTABF.",
    "MOSTLYAI": "MOSTLYAI",
    "GRETEL_ACTGAN": "GRE-ACTGAN",
    "GRETEL_LSTM": "GRE-LSTM",
    "CLAVADDPM": "CLAVADDPM",
}

baselines = {"walmart": 7697, "rossmann": 345, "airbnb": 0.5}

dataset_names = {"walmart": "Walmart", "rossmann": "Rossmann", "airbnb": "AirBnB"}


def df_to_latex(df):
    latex_table = df.to_latex(index=False, escape=False, column_format="clrrr")
    # replace all white spaces with one space
    table_latex = re.sub(" +", " ", latex_table)
    table_latex = table_latex.replace("\\\\\n\\multirow", "\\\\\n\\hline\n\\multirow")
    table_latex = (
        table_latex.replace(".00\\", "\\")
        .replace(".000\\", "\\")
        .replace(".000$", "$")
        .replace(".00$", "$")
    )
    return table_latex


def add_utility_results(df, dataset, data, round_score=False):
    model = "xgboost"
    score = np.abs(data[dataset]["SDV"][model]["real_score"])
    score_se = data[dataset]["SDV"][model]["real_score_se"]
    if round_score:
        score = np.round(score).astype(int)
        if score_se > 1:
            score_se = np.round(score_se).astype(int)
        score_se = np.round(score_se, 1)

    header_row = pd.DataFrame(
        {
            "Dataset": f"\multirow{{7}}{{*}}{{\\parbox{{1.3cm}}{{{dataset_names[dataset]}}}}}",
            "Method": "Real Data",
            "Score": f"${score:.2f}\pm{score_se:.3f}$ ({baselines[dataset]})",
            "Model Selection": "-",
            "Feature Selection": "-",
        },
        index=[0],
    )
    # add the row without append
    df = pd.concat([df, header_row], ignore_index=True)

    for method in data[dataset].keys():
        score = np.abs(data[dataset][method][model]["synthetic_score"])
        score_se = data[dataset][method][model]["synthetic_score_se"]
        if round_score:
            score = np.round(score).astype(int)
            if score_se > 1:
                score_se = np.round(score_se).astype(int)
            score_se = np.round(score_se, 1)

        model_selection = data[dataset][method]["spearman_mean"]
        model_selection_se = data[dataset][method]["spearman_se"]
        feature_selection = data[dataset][method]["feature_importance_spearman_mean"]
        feature_selection_se = data[dataset][method]["feature_importance_spearman_se"]

        new_row = pd.DataFrame(
            {
                "Dataset": "",
                "Method": method_names_dict[method],
                "Score": f"${score:.2f}\pm{score_se:.3f}$",
                "Model Selection": f"${model_selection:.2f}\pm{model_selection_se:.2f}$",
                "Feature Selection": f"${feature_selection:.2f}\pm{feature_selection_se:.2f}$",
            },
            index=[0],
        )
        df = pd.concat([df, new_row], ignore_index=True)
    return df


def add_rank_results(df, dataset, data, feature_importance=False):
    methods = list(data[dataset].keys())
    method = methods[0]
    if feature_importance:
        key_spearman = "feature_importance_spearman"
        key_kendall = "feature_importance_tau"
        key_weighted = "feature_importance_weighted"
    else:
        key_spearman = "spearman"
        key_kendall = "kendall"
        key_weighted = "weighted"
    spearman = data[dataset][method][f"{key_spearman}_mean"]
    spearman_se = data[dataset][method][f"{key_spearman}_se"]
    kendall_Kendall = data[dataset][method][f"{key_kendall}_mean"]
    kendall_Kendall_se = data[dataset][method][f"{key_kendall}_se"]
    weighted = data[dataset][method][f"{key_weighted}_mean"]
    weighted_se = data[dataset][method][f"{key_weighted}_se"]

    header_row = pd.DataFrame(
        {
            "Dataset": f"\multirow{{7}}{{*}}{{\\parbox{{1.3cm}}{{{dataset_names[dataset]}}}}}",
            "Method": method_names_dict[method],
            "Spearman": f"${spearman:.2f}\pm{spearman_se:.2f}$",
            "Kendall": f"${kendall_Kendall:.2f}\pm{kendall_Kendall_se:.2f}$",
            "Weighted": f"${weighted:.2f}\pm{weighted_se:.2f}$",
        },
        index=[0],
    )
    df = pd.concat([df, header_row], ignore_index=True)

    for method in methods[1:]:
        spearman = data[dataset][method][f"{key_spearman}_mean"]
        spearman_se = data[dataset][method][f"{key_spearman}_se"]
        kendall_Kendall = data[dataset][method][f"{key_kendall}_mean"]
        kendall_Kendall_se = data[dataset][method][f"{key_kendall}_se"]
        weighted = data[dataset][method][f"{key_weighted}_mean"]
        weighted_se = data[dataset][method][f"{key_weighted}_se"]

        new_row = pd.DataFrame(
            {
                "Dataset": "",
                "Method": method_names_dict[method],
                "Spearman": f"${spearman:.2f}\pm{spearman_se:.2f}$",
                "Kendall": f"${kendall_Kendall:.2f}\pm{kendall_Kendall_se:.2f}$",
                "Weighted": f"${weighted:.2f}\pm{weighted_se:.2f}$",
            },
            index=[0],
        )
        df = pd.concat([df, new_row], ignore_index=True)
    return df


df = pd.DataFrame(
    columns=["Dataset", "Method", "Score", "Model Selection", "Feature Selection"]
)
df_model = pd.DataFrame(
    columns=["Dataset", "Method", "Spearman", "Kendall", "Weighted"]
)
df_feature = pd.DataFrame(
    columns=["Dataset", "Method", "Spearman", "Kendall", "Weighted"]
)

for dataset in ["airbnb", "rossmann", "walmart"]:
    with open(f"results/mle_{dataset}_{run}_0.json") as f:
        results = json.load(f)
    round_score = dataset != "airbnb"
    df = add_utility_results(df, dataset, results, round_score)
    df_model = add_rank_results(df_model, dataset, results)
    df_feature = add_rank_results(df_feature, dataset, results, feature_importance=True)

# Save the tables


print("Saving Table 2 (best results were bolded manually).")
table_latex = df_to_latex(df)
with open("results/tables/table2.tex", "w") as f:
    f.write(table_latex)

print("Saving Table 7 (best results were bolded manually).")
table_latex = df_to_latex(df_model)
with open("results/tables/table7.tex", "w") as f:
    f.write(table_latex)

print("Saving Table 8 (best results were bolded manually).")
table_latex = df_to_latex(df_feature)
with open("results/tables/table8.tex", "w") as f:
    f.write(table_latex)



# -*- coding: utf-8 -*-
# configure logging for the library with a null handler (nothing is printed by default). See
# http://docs.python-guide.org/en/latest/writing/logging/

"""Top-level package for SDV."""

__author__ = """MIT Data To AI Lab"""
__email__ = 'dailabmit@gmail.com'
__version__ = '0.13.1'

from rctgan import constraints, evaluation, metadata, relational, tabular
from rctgan.demo import get_available_demos, load_demo
from rctgan.metadata import Metadata, Table
from rctgan.rctgan import RCTGAN
from rctgan.rdt2.transformers.base import BaseTransformer
from rctgan.rdt2.transformers.boolean import BinaryEncoder
from rctgan.rdt2.transformers.categorical import FrequencyEncoder, LabelEncoder, OneHotEncoder
from rctgan.rdt2.transformers.datetime import OptimizedTimestampEncoder, UnixTimestampEncoder
from rctgan.rdt2.transformers.null import NullTransformer
from rctgan.rdt2.transformers.numerical import ClusterBasedNormalizer, FloatFormatter, GaussianNormalizer
from rctgan.rdt2.transformers.pii.anonymizer import AnonymizedFaker

__all__ = (
    'rdt2',
    'ctganpc',
    'demo',
    'constraints',
    'evaluation',
    'metadata',
    'relational',
    'tabular',
    'get_available_demos',
    'load_demo',
    'Metadata',
    'Table',
    'RCTGAN',
)
